/*
	This little utility dumps the EEPROM of a
	National Instruments AT-MIO E series card and
	prints the information contained therein in
	a human-readable form.
	
	Information here is taken from the AT-MIO E
	series Register-Level Programmer Manual.  It
	doesn't give much information about what the
	CALDACs actually do, but it shouldn't be
	difficult to figure out what they do.
*/

#include <stdio.h>
#include <comedilib.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <errno.h>
#include <getopt.h>
#include <ctype.h>

/* This structure is taken directly from module/ni_E.c */
struct ni_boardtype_struct{
	int boardcode;
	char *name;
	int n_adchan;
	int adbits;
	int alwaysdither;
	int limitedgain;
	int caldactype;
	int eepromlayout;
};

static struct ni_boardtype_struct ni_boardtypes[]={
	{44,"at-mio-16e-1"  ,16,12,0,0,1,1 },
	{25,"at-mio-16e-2"  ,16,12,0,0,1,1 },
	{36,"at-mio-16e-10" ,16,12,0,0,1,2 },
	{37,"at-mio-16de-10",16,12,0,0,1,2 },
	{38,"at-mio-64e-3"  ,64,12,0,0,1,1 },
	{39,"at-mio-16xe-50",16,16,1,1,2,3 }
};
#define N_BOARDTYPES 6

void print_calib1(unsigned char *calib);
void print_calib2(unsigned char *calib);
void print_calib3(unsigned char *calib);
int get_boardtype(int eeprom511);
void do_rawdump(char *eeprom);

int main(int argc,char *argv[])
{
	int it;
	char *fn=NULL;
	comedi_memcpy mc;
	unsigned char eeprom[512];
	int boardtype;
	int raw=0;
	int c;

	fn="/dev/comedi0";
	while(1){
		c=getopt(argc,argv,"rf");
		if(c==-1)break;
		switch(c){
		case 'r':
			raw=1;
			break;
		case 'f':
			fn=argv[optind];
			break;
		default:
			printf("bad option\n");
			exit(1);
		}
	}

	it=open(fn,O_RDWR);
	if(it<0){
		perror(fn);
		exit(0);
	}

	mc.local=eeprom;
	mc.remote=0;
	mc.nbytes=512;
	if(ioctl(it,COMEDI_COPY_FROM,&mc)<0){
		perror("possibly not NI at-mio E series board?");
		exit(1);
	}
		
	if(raw){
		do_rawdump(eeprom);
		exit(0);
	}

	printf("Board code:     0x%02x\n",eeprom[511]);
	printf("Revision:       0x%02x\n",eeprom[510]);
	printf("Sub-revision:   0x%02x\n",eeprom[509]);
	printf("\nFactory Calibration:  (%d/%d/%d)\n",eeprom[507],eeprom[506],eeprom[508]);
	
	boardtype=get_boardtype(eeprom[511]);
	switch(ni_boardtypes[boardtype].eepromlayout){
	case 1:
		print_calib1(eeprom+409);
		printf("\nUser Calibration 1:\n");
		print_calib1(eeprom+371+1-18);
		printf("\nUser Calibration 2:\n");
		print_calib1(eeprom+371+1-18*2);
		printf("\nUser Calibration 3:\n");
		print_calib1(eeprom+371+1-18*3);
		printf("\nUser Calibration 4:\n");
		print_calib1(eeprom+371+1-18*3);
		printf("\nUser Calibration 5:\n");
		print_calib1(eeprom+371+1-18*3);
		break;
	case 2:
		print_calib2(eeprom+406);
		printf("\nUser Calibration 1:\n");
		print_calib2(eeprom+371+1-19);
		printf("\nUser Calibration 2:\n");
		print_calib2(eeprom+371+1-19*2);
		printf("\nUser Calibration 3:\n");
		print_calib2(eeprom+371+1-19*3);
		printf("\nUser Calibration 4:\n");
		print_calib2(eeprom+371+1-19*3);
		printf("\nUser Calibration 5:\n");
		print_calib2(eeprom+371+1-19*3);
		break;
	case 3:
		print_calib2(eeprom+423);
		printf("\nUser Calibration 1:\n");
		print_calib2(eeprom+371+1-16);
		printf("\nUser Calibration 2:\n");
		print_calib2(eeprom+371+1-16*2);
		printf("\nUser Calibration 3:\n");
		print_calib2(eeprom+371+1-16*3);
		printf("\nUser Calibration 4:\n");
		print_calib2(eeprom+371+1-16*3);
		printf("\nUser Calibration 5:\n");
		print_calib2(eeprom+371+1-16*3);
		break;
	default:
		printf("unknown EEPROM layout\n");
		break;
	}
	
	return 0;
}

int get_boardtype(int eeprom511)
{
	int i;
	
	for(i=0;i<N_BOARDTYPES;i++){
		if(eeprom511==ni_boardtypes[i].boardcode)return i;
	}
	printf("Unknown board code.\n");
	return 0;
}

void print_calib1(unsigned char *calib)
{
	/*
	   this is the EEPROM calibration layout for
	   at-mio-16e-1, 16e-2, and 64e-3
	
	   one can only wonder why there are gratuitous
	   changes between these layouts...
	*/
	printf("  reference: 0x%02x%02x\n",calib[17],calib[16]);
	printf("  analog output (unipolar):\n");
	printf("    CALDAC 5:   0x%02x\n",calib[ 5]);
	printf("    CALDAC 6:   0x%02x\n",calib[ 3]);
	printf("    CALDAC 7:   0x%02x\n",calib[ 4]);
	printf("    CALDAC 8:   0x%02x\n",calib[ 2]);
	printf("    CALDAC 9:   0x%02x\n",calib[ 0]);
	printf("    CALDAC 10:  0x%02x\n",calib[ 1]);
	printf("  analog output (bipolar):\n");
	printf("    CALDAC 5:   0x%02x\n",calib[11]);
	printf("    CALDAC 6:   0x%02x\n",calib[ 9]);
	printf("    CALDAC 7:   0x%02x\n",calib[10]);
	printf("    CALDAC 8:   0x%02x\n",calib[ 8]);
	printf("    CALDAC 9:   0x%02x\n",calib[ 6]);
	printf("    CALDAC 10:  0x%02x\n",calib[ 7]);
	printf("  analog input:\n");
	printf("    CALDAC 1:   0x%02x\n",calib[14]);
	printf("    CALDAC 2:   0x%02x\n",calib[12]);
	printf("    CALDAC 3:   0x%02x\n",calib[13]);
	printf("    CALDAC 4:   0x%02x\n",calib[15]);
}

void print_calib2(unsigned char *calib)
{
	/*
	   this is the EEPROM calibration layout for
	   at-mio-16e-10, 16de-10
	*/
	printf("  reference: 0x%02x%02x\n",calib[18],calib[17]);
	printf("  analog output (unipolar):\n");
	printf("    CALDAC 5:   0x%02x\n",calib[ 5]);
	printf("    CALDAC 6:   0x%02x\n",calib[ 3]);
	printf("    CALDAC 7:   0x%02x\n",calib[ 4]);
	printf("    CALDAC 8:   0x%02x\n",calib[ 2]);
	printf("    CALDAC 9:   0x%02x\n",calib[ 0]);
	printf("    CALDAC 10:  0x%02x\n",calib[ 1]);
	printf("  analog output (bipolar):\n");
	printf("    CALDAC 5:   0x%02x\n",calib[11]);
	printf("    CALDAC 6:   0x%02x\n",calib[ 9]);
	printf("    CALDAC 7:   0x%02x\n",calib[10]);
	printf("    CALDAC 8:   0x%02x\n",calib[ 8]);
	printf("    CALDAC 9:   0x%02x\n",calib[ 6]);
	printf("    CALDAC 10:  0x%02x\n",calib[ 7]);
	printf("  analog input:\n");
	printf("    CALDAC 1:   0x%02x\n",calib[14]);
	printf("    CALDAC 2:   0x%02x\n",calib[12]);
	printf("    CALDAC 3:   0x%02x\n",calib[13]);
	printf("    CALDAC 4:   0x%02x\n",calib[16]);
	printf("    CALDAC 11:  0x%02x\n",calib[15]);
}

void print_calib3(unsigned char *calib)
{
	/*
	   this is the EEPROM calibration layout for
	   at-mio-16xe-50
	*/
	printf("  reference: 0x%02x%02x\n",calib[15],calib[14]);
	printf("  analog output:\n");
	printf("    CALDAC 4:   0x%02x\n",calib[ 2]);
	printf("    CALDAC 5:   0x%02x\n",calib[ 0]);
	printf("    CALDAC 6:   0x%02x\n",calib[ 3]);
	printf("    CALDAC 7:   0x%02x\n",calib[ 1]);
	printf("  analog input (unipolar):\n");
	printf("    CALDAC 0:   0x%02x\n",calib[ 5]);
	printf("    CALDAC 1:   0x%02x\n",calib[ 4]);
	printf("    CALDAC 2:   0x%02x\n",calib[ 6]);
	printf("    CALDAC 8:   0x%02x%02x\n",calib[ 8],calib[7]);
	printf("  analog input (unipolar):\n");
	printf("    CALDAC 0:   0x%02x\n",calib[10]);
	printf("    CALDAC 1:   0x%02x\n",calib[ 9]);
	printf("    CALDAC 2:   0x%02x\n",calib[11]);
	printf("    CALDAC 8:   0x%02x%02x\n",calib[13],calib[12]);
}

void do_rawdump(char *eeprom)
{
	int i,j,c;
	
	printf("NI AT-MIO E series EEPROM dump\n");
	for(i=0;i<32;i++){
		for(j=0;j<16;j++){
			printf("%02x",eeprom[i*16+j]&0xff);
		}
		printf("  ");
		for(j=0;j<16;j++){
			c=eeprom[i*16+j]&0xff;
			printf("%c",isprint(c)?c:'.');
		}
		printf("\n");
	}
}

