/*
    lib/comedi.c
    comedi library routines

    COMEDI - Linux Control and Measurement Device Interface
    Copyright (C) 1997-8 David A. Schleef <ds@stm.lbl.gov>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <errno.h>
#include <comedi.h>
#include <string.h>

#define inline 

#include <comedilib.h>

unsigned int comedi_version_code=COMEDI_VERSION_CODE;


int comedi_put(int fd,int chan,int data)
{
	comedi_trig it;

	it.chan=chan;
	it.io=COMEDI_IO_OUTPUT;
	it.n=0;
	it.trigsrc=COMEDI_TRIGNOW;
	it.data=data;

	return ioctl(fd,COMEDI_TRIG,&it);
}

int comedi_masked_put(int fd,int chan,int mask,int data)
{
	comedi_trig it;

	it.chan=chan;
	it.io=COMEDI_IO_OUTPUT;
	it.n=mask;
	it.trigsrc=COMEDI_TRIGNOW;
	it.data=data;

	return ioctl(fd,COMEDI_TRIG,&it);
}
	
int comedi_get(int fd,int chan,int *data)
{
	comedi_trig it;
	int ret;
	
	it.chan=chan;
	it.io=COMEDI_IO_INPUT;
	it.n=1;
	it.trigsrc=COMEDI_TRIGNOW;

	ret=ioctl(fd,COMEDI_TRIG,&it);

	*data=it.data;
	
	return ret;
}

int comedi_getparam(int fd,int chan,int pnum,int *flags,int *pval)
{
	comedi_param it;
	int ret;
	
	it.chan=chan;
	it.pnum=pnum;
	ret=ioctl(fd,COMEDI_GETPARAM,&it);
	*flags=it.flags;
	*pval=it.pval;
	
	return ret;
}

int comedi_setparam(int fd,int chan,int pnum,int pval)
{
	comedi_param it;
	int ret;

	it.chan=chan;
	it.pnum=pnum;
	it.pval=pval;
	ret=ioctl(fd,COMEDI_SETPARAM,&it);
	
	return ret;
}

int comedi_getinfo(int fd,comedi_info *it)
{
	int ret;

	ret=ioctl(fd,COMEDI_GETINFO,it);

	return ret;
}

/*
int comedi_init(char *s)
{
	return open(s,O_RDWR);
}

void comedi_reset(int fd)
{
	ioctl(fd,COMEDI_RESET,0);
}
*/

/*
int comedi_memcpy_to(int fd,void *src,int dest,int size)
{
	comedi_copy it;

	it.src=(int)src;
	it.dest=dest;
	it.nbytes=size;

	return ioctl(fd,COMEDI_COPY_TO,&it);
}

int comedi_memset(int fd,int dest,int size,int val)
{
	comedi_copy it;

	it.src=val;
	it.dest=dest;
	it.nbytes=size;

	return ioctl(fd,COMEDI_MEMSET,&it);
}


int comedi_memcpy_from(int fd,int src,void *dest,int size)
{
	comedi_copy it;

	it.src=src;
	it.dest=(int)dest;
	it.nbytes=size;

	return ioctl(fd,COMEDI_COPY_FROM,&it);
}
*/

/*
int comedi_exec(int fd,int pc)
{
	return ioctl(fd,COMEDI_EXEC,pc);
}
*/

static double dt2821_min[]={ -10, -5, -2.5, -1.25 };
static double dt2821_max[]={ 10, 5, 2.5, 1.25 };

static double ni_E_min[]={ -10, -5, -2.5, -1, -0.5, -0.25, -0.1, -0.05,
	0,0,0,0,0,0,0,0 };
static double ni_E_max[]={ 10, 5, 2.5, 1, 0.5, 0.25, 0.1, 0.05,
	20, 10, 5, 2, 1, 5, 0.2, 0.1 };

static double pcl711b_min[]={ -5, -2.5, -1.25, -0.625, -0.3125 };
static double pcl711b_max[]={ 5, 2.5, 1.25, 0.625, 0.3125 };

static double rti800_min[]={ 0, 0, 0, 0 };
static double rti800_max[]={ 10, 1, 0.1, 0.02 };

int comedi_voltage_range(int fd,unsigned int chan,unsigned int range,
	comedi_voltage_info *vinfo)
{
	comedi_info info;
	int ret;
	
	ret=ioctl(fd,COMEDI_GETINFO,&info);
	if(ret<0)return ret;
	
	vinfo->maxdata=0;
	vinfo->offset=0;
	vinfo->max=0;
	vinfo->delta=0;

	if(!strcmp(info.board_name,"dt2821")){
		if(chan>15)return 0;
		if(range>3)return -EINVAL;
		vinfo->maxdata=4095;
		vinfo->offset=dt2821_min[range];
		vinfo->max=dt2821_max[range];
		vinfo->delta=(vinfo->max-vinfo->offset)/4095;
		return 0;
	}
	
	if(!strcmp(info.board_name,"at-mio-16e-1") ||
	   !strcmp(info.board_name,"at-mio-16e-2") ||
	   !strcmp(info.board_name,"at-mio-16e-10") ||
	   !strcmp(info.board_name,"at-mio-16de-10")){
		if(chan<16){
			if(range>15)return -EINVAL;
			vinfo->maxdata=4095;
			vinfo->offset=ni_E_min[range];
			vinfo->max=ni_E_max[range];
			vinfo->delta=(vinfo->max-vinfo->offset)/4095;
			return 0;
		}
		if(chan<18){
			vinfo->maxdata=4095;
			if(range==0){
				vinfo->offset=0;
				vinfo->max=10.0;
			}else{
				vinfo->offset=-10.0;
				vinfo->max=10.0;
			}
			vinfo->delta=(vinfo->max-vinfo->offset)/4095;
			return 0;
		}
		return -EINVAL;
	}
	if(!strcmp(info.board_name,"at-mio-64e-3")){
		if(chan<64){
			if(range>15)return -EINVAL;
			vinfo->maxdata=4095;
			vinfo->offset=ni_E_min[range];
			vinfo->max=ni_E_max[range];
			vinfo->delta=(vinfo->max-vinfo->offset)/4095;
			return 0;
		}
		if(chan<66){
			vinfo->maxdata=4095;
			if(range==0){
				vinfo->offset=0;
				vinfo->max=10.0;
			}else{
				vinfo->offset=-10.0;
				vinfo->max=10.0;
			}
			vinfo->delta=(vinfo->max-vinfo->offset)/4095;
			return 0;
		}
	}
	if(!strcmp(info.board_name,"pcl711")){
		if(chan<8){
			vinfo->maxdata=4095;
			vinfo->offset=-5;
			vinfo->max=5;
			vinfo->delta=(vinfo->max-vinfo->offset)/4095;
			return 0;
		}
		if(chan<9){
			vinfo->maxdata=4095;
			vinfo->offset=0;
			vinfo->max=10;
			vinfo->delta=(vinfo->max-vinfo->offset)/4095;
			return 0;
		}
		return -EINVAL;
	}
	if(!strcmp(info.board_name,"pcl711b")){
		if(chan<8){
			if(range>4)return -EINVAL;
			vinfo->maxdata=4095;
			vinfo->offset=pcl711b_min[range];
			vinfo->max=pcl711b_max[range];
			vinfo->delta=(vinfo->max-vinfo->offset)/4095;
			return 0;
		}
		if(chan<9){
			vinfo->maxdata=4095;
			vinfo->offset=0;
			vinfo->max=10;
			vinfo->delta=(vinfo->max-vinfo->offset)/4095;
			return 0;
		}
		return -EINVAL;
	}
	if(!strcmp(info.board_name,"dt2814")){
		vinfo->maxdata=4095;
		vinfo->offset=-5;
		vinfo->max=5;
		vinfo->delta=(vinfo->max-vinfo->offset)/4095;
		return 0;
	}
	if(!strcmp(info.board_name,"rti800")){
		if(chan<8){
			if(range>3)return -EINVAL;
			vinfo->maxdata=4095;
			vinfo->offset=rti800_min[range];
			vinfo->max=rti800_max[range];
			vinfo->delta=(vinfo->max-vinfo->offset)/4095;
			return 0;
		}
	}
	return -EINVAL;
}

int comedi_get_timer(int fd,unsigned int chan,double *freq,
	unsigned int *trigvar)
{
	comedi_info info;
	int ret;
	
	ret=ioctl(fd,COMEDI_GETINFO,&info);
	if(ret<0)return ret;
	
	if(*freq<0)return -1;
	if(!strncmp(info.board_name,"dt282",5)){
		int divider,prescaler;
		double basefreq=4e6;
		
		divider=floor(4e6/(*freq));
		prescaler=0;
		while(divider>255){
			prescaler++;
			divider>>=1;
			basefreq/=2;
		}
		if(prescaler==1){
			prescaler++;
			divider>>=1;
			basefreq/=2;
		}
		if(prescaler>=16)return -1;
		*trigvar=(prescaler<<8)|(255-divider);
		*freq=basefreq/divider;
		
		return 0;
	}
	if(!strcmp(info.board_name,"dt2814")){
		double f;
		int  i;
		
		f=1e5;
		for(i=0;i<8;i++){
			if(f-*freq<*freq-f/10){
				*trigvar=i;
				*freq=f;
				return 0;
			}
			f/=10;
		}
		*trigvar=i;
		*freq=f;
		return 0;
	}
	if(!strncmp(info.board_name,"at-mio",6)){
		unsigned int divider;
		
		divider=floor(20e6/(*freq));
		*freq=20e6/divider;
		*trigvar=divider-1;
		
		return 0;
	}
	return -1;
}

