/*
    lib/range.c
    comedi library routines for voltage ranges

    COMEDI - Linux Control and Measurement Device Interface
    Copyright (C) 1997-8 David A. Schleef <ds@stm.lbl.gov>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <errno.h>
#include <comedi.h>
#include <string.h>

#include <comedilib.h>


/* general */

static comedi_range bipolar10[]={ {-10,10}, };
static comedi_range bipolar5[]={ {-5,5}, };
static comedi_range bipolar2_5[]={ {-2.5,2.5}, };
static comedi_range unipolar10[]={ {0,10}, };
static comedi_range unipolar5[]={ {0,5}, };
static comedi_range unipolar2_5[]={ {0,2.5}, };

/* dt282x */

static comedi_range dt282x_ai_lo_bipolar[]={
	{-10,10},
	{-5,5},
	{-2.5,2.5},
	{-1.25,1.25}
};

static comedi_range dt282x_ai_lo_unipolar[]={
	{0,10},
	{0,5},
	{0,2.5},
	{0,1.25},
};

static comedi_range dt282x_ai_5_bipolar[]={
	{-5,5},
	{-2.5,2.5},
	{-1.25,1.25},
	{-0.625,0.625},
};

static comedi_range dt282x_ai_5_unipolar[]={
	{0,5},
	{0,2.5},
	{0,1.25},
	{0,0.625},
};

static comedi_range dt282x_ai_hi_bipolar[]={
	{-10,10},
	{-1,1},
	{-0.1,0.1},
	{-0.02,0.02}
};

static comedi_range dt282x_ai_hi_unipolar[]={
	{0,10},
	{0,1},
	{0,0.1},
	{0,0.02},
};

/* ni_E */

static comedi_range ni_E_ai_ranges[]={
	{-10,10},
	{-5,5},
	{-2.5,2.5},
	{-1,1},
	{-0.5,0.5},
	{-0.25,0.25},
	{-0.1,0.1},
	{-0.05,0.05},
	{0,20},
	{0,10},
	{0,5},
	{0,2},
	{0,1},
	{0,0.5},
	{0,0.2},
	{0,0.1}
};

static comedi_range ni_E_ai_limited[]={
	{-5,5},
	{-2.5,2.5},
	{-0.5,0.5},
	{-0.05,0.05},
	{0,10},
	{0,5},
	{0,1},
	{0,0.1}
};

/* pcl711b */

static comedi_range pcl711b_ai[]={
	{-5,5},
	{-2.5,2.5},
	{-1.25,1.25},
	{-0.625,0.625},
	{-0.3125,0.3125}
};

/* rti800 */

static comedi_range rti800_ranges[]={
	{0,10},
	{0,1},
	{0,0.1},
	{0,0.02}
};





static comedi_range_info vinfos[]={
	{},
	{ 1, unipolar10 },
	{ 1, unipolar5 },
	{ 1, bipolar10 },
	{ 1, bipolar5 },
	{ 4, dt282x_ai_lo_bipolar },
	{ 4, dt282x_ai_lo_unipolar },
	{ 4, dt282x_ai_hi_bipolar },
	{ 4, dt282x_ai_hi_unipolar },
	{ 16, ni_E_ai_ranges },
	{ 1, unipolar2_5 },
	{ 1, bipolar2_5 },
	{ 4, dt282x_ai_5_bipolar },
	{ 4, dt282x_ai_5_unipolar },
	{ 5, pcl711b_ai },
	{ 8, ni_E_ai_limited },
};
#define N_VINFOS (sizeof(vinfos)/sizeof(comedi_range_info))


int comedi_get_range_info(int fd,unsigned int chan,comedi_range_info **vinfo)
{
	int ret;
	int flags,range_type=0;
	
	ret=comedi_getparam(fd,chan,COMEDI_RANGE_TYPE,&flags,&range_type);
	if(ret<0)
		return ret;
	
	if(range_type==0 || range_type>=N_VINFOS)
		return -1;

	*vinfo=vinfos+range_type;
	
	return 0;
}

