/*
 * Driver for Data Translation DT2811
 * David A. Schleef
 */

#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/mm.h>
#include <linux/malloc.h>
#include <linux/errno.h>
#include <linux/ioport.h>
#include <linux/timer.h>
#include <linux/delay.h>
#include <linux/interrupt.h>
#include <linux/timex.h>
#include <asm/io.h>
#include <comedi_module.h>

/*

0x00    ADCSR R/W  A/D Control/Status Register
  bit 7 - (R) 1 indicates A/D conversion done
          reading ADDAT clears bit
          (W) ignored
  bit 6 - (R) 1 indicates A/D error
          (W) ignored
  bit 5 - (R) 1 indicates A/D busy, cleared at end
          of conversion
          (W) ignored
  bit 4 - (R) 0
          (W) 
  bit 3 - (R) 0
  bit 2 - (R/W) 1 indicates interrupts enabled
  bits 1,0 - (R/W) mode bits
     00  single conversion on ADGCR load
     01  continuous conversion, internal clock,
         (clock enabled on ADGCR load)
     10  continuous conversion, internal clock,
         external trigger
     11  continuous conversion, external clock,
         external trigger

0x01    ADGCR R/W A/D Gain/Channel Register
  bit 6,7 - (R/W) gain select
    00  gain=1, both PGH, PGL models
    01  gain=2 PGH, 10 PGL
    10  gain=4 PGH, 100 PGL
    11  gain=8 PGH, 500 PGL
  bit 4,5 - reserved
  bit 3-0 - (R/W) channel select
    channel number from 0-15

0x02,0x03 (R) ADDAT A/D Data Register
          (W) DADAT0 D/A Data Register 0
  0x02 low byte
  0x03 high byte

0x04,0x05 (W) DADAT0 D/A Data Register 1

0x06 (R) DIO0 Digital Input Port 0
     (W) DIO1 Digital Output Port 1

0x07 TMRCTR (R/W) Timer/Counter Register
  bits 6,7 - reserved
  bits 5-3 - Timer frequency control (mantissa)
    543  divisor  freqency (kHz)
    000  1        600
    001  10       60
    010  2        300
    011  3        200
    100  4        150
    101  5        120
    110  6        100
    111  12       50
  bits 2-0 - Timer frequency control (exponent)
    210  multiply divisor/divide frequency by
    000  1
    001  10
    010  100
    011  1000
    100  10000
    101  100000
    110  1000000
    111  10000000

*/

#define DT2811_SIZE 8

#define DT2811_ADCSR 0
#define DT2811_ADGCR 1
#define DT2811_ADDATLO 2
#define DT2811_ADDATHI 3
#define DT2811_DADAT0LO 2
#define DT2811_DADAT0HI 3
#define DT2811_DADAT1LO 4
#define DT2811_DADAT1HI 5
#define DT2811_DIO 6
#define DT2811_TMRCTR 7

/*
 * flags
 */

/* ADCSR */

#define DT2811_ADDONE   0x80
#define DT2811_ADERROR  0x40
#define DT2811_ADBUSY   0x20
#define DT2811_CLRERROR 0x10
#define DT2811_INTENB   0x04
#define DT2811_ADMODE   0x03

comedi_devinit dt2811_init;
static int dt2811_rem(comedi_device *dev);
static int dt2811_ad(comedi_device *,comedi_trig *);
static int dt2811_da(comedi_device *,comedi_trig *);
static int dt2811_di(comedi_device *,comedi_trig *);
static int dt2811_do(comedi_device *,comedi_trig *);

#if 0
static void dt2811_interrupt(int irq,void *dev,struct pt_regs * regs)
{
	int lo,hi,usec;
	comedi_device *d=dev;
	struct timeval it;
	comedi_sample pt;


switch(d->i_admode){
case COMEDI_MDEMAND:
	lo=inb(d->iobase+DT2811_ADDATLO);
	hi=inb(d->iobase+DT2811_ADDATHI);

	pt.data=lo+0x100*hi;
	pt.chan=d->curadchan;
	pt.time=trigtime.tv_usec;
	copy_to_buffer(&(d->rdbuf),&pt,sizeof(comedi_sample));

	if(d->ntrig>0){
		d->ntrig--;
		outb(d->curadchan,d->iobase+DT2811_ADGCR);
		do_gettimeofday(&trigtime);
	}else{
		wake_up_interruptible(&d->rdwait);
	}
	break;
case COMEDI_MCONTS:
	lo=inb(d->iobase+DT2811_ADDATLO);
	hi=inb(d->iobase+DT2811_ADDATHI);
	do_gettimeofday(&it);
	if(d->ntrig>0){
		/*do_gettimeofday(&it);*/
		d->ntrig--;
		/*pt.data=lo+0x100*hi;*/
		pt.data=it.tv_usec;
		pt.chan=d->curadchan;
		pt.time=trigtime.tv_usec;
		copy_to_buffer(&(d->rdbuf),&pt,sizeof(comedi_sample));
		wake_up_interruptible(&d->rdwait);
	}
	if(inb(d->iobase+DT2811_ADCSR)&DT2811_ADERROR){
		printk("dt2811: ad error %d %d\n",it.tv_sec,it.tv_usec);
		outb(DT2811_CLRERROR|DT2811_INTENB|1,d->iobase+DT2811_ADCSR);
	}
	trigtime.tv_usec+=10000;
	if(trigtime.tv_usec>=1000000){
		trigtime.tv_usec-=1000000;
		trigtime.tv_sec++;
	}
	break;
}
}
#endif

int dt2811_init(comedi_device *dev,comedi_devconfig *it)
{
	int i;

	if(strcmp("dt2811",it->name))
		return 0;

	dev->iobase=0x240;
	if(it->options[0])dev->iobase=it->options[0];
	
	printk("comedi%d: dt2811: base=0x%04x ",dev->minor,dev->iobase);
	
	if(check_region(dev->iobase,DT2811_SIZE)<0){
		printk("I/O port conflict\n");
		return -EIO;
	}
	request_region(dev->iobase,DT2811_SIZE,"dt2811");
	dev->driver_name="dt2811";
	dev->board_name="dt2811";
	dev->iosize=DT2811_SIZE;

	outb(0,dev->iobase+DT2811_ADCSR);
	udelay(100);
	i=inb(dev->iobase+DT2811_ADDATLO);
	i=inb(dev->iobase+DT2811_ADDATHI);

#if 0
	save_flags(flags);
	sti();
	irqs=probe_irq_on();

	outb(DT2811_CLRERROR|DT2811_INTENB,dev->iobase+DT2811_ADCSR);
	outb(0,dev->iobase+DT2811_ADGCR);

	udelay(100);

	irq=probe_irq_off(irqs);
	restore_flags(flags);

	/*outb(DT2811_CLRERROR|DT2811_INTENB,dev->iobase+DT2811_ADCSR);*/

	irq=0;
	dev->irq=0;
	if(irq>0){
		i=inb(dev->iobase+DT2811_ADDATLO);
		i=inb(dev->iobase+DT2811_ADDATHI);
		printk("(irq = %d)\n",irq);
		/* request_irq(irq,dt2811_interrupt,0,"dt2811",dev); */
		dev->irq=irq;
	}else if(irq==0){
		printk("(no irq)\n");
	}else{
		printk("( multiple irq's -- this is bad! )\n");
	}
#endif
	dev->irq=0;

	dev->chan=20;
	dev->chinfo=kmalloc(dev->chan*sizeof(comedi_chinfo),GFP_KERNEL);
	if(!dev->chinfo)return -ENOMEM;

	memset(dev->chinfo,0,dev->chan*sizeof(comedi_chinfo));
	for(i=0;i<16;i++){
		dev->chinfo[i].trig=dt2811_ad;
		dev->chinfo[i].sp=NULL;
		addparam(&dev->chinfo[i].paramlist,COMEDI_BITS,COMEDI_RDONLY,
			12);
		addparam(&dev->chinfo[i].paramlist,COMEDI_MAXDATA,COMEDI_RDONLY,
			0xfff);
		addparam(&dev->chinfo[i].paramlist,COMEDI_AD,COMEDI_RDONLY,
			COMEDI_AD_ANALOG);
		addparam(&dev->chinfo[i].paramlist,COMEDI_IO,COMEDI_RDONLY,
			COMEDI_IO_INPUT);
		addparam(&dev->chinfo[i].paramlist,COMEDI_LASTDATA,
			COMEDI_RDONLY,0);
		dev->chinfo[i].lastdata=dev->chinfo[i].paramlist;
	}
	
	dev->chinfo[16].trig=dt2811_da;
	dev->chinfo[16].sp=NULL;
	addparam(&dev->chinfo[16].paramlist,COMEDI_BITS,COMEDI_RDONLY,
		12);
	addparam(&dev->chinfo[16].paramlist,COMEDI_MAXDATA,COMEDI_RDONLY,
		0xfff);
	addparam(&dev->chinfo[16].paramlist,COMEDI_AD,COMEDI_RDONLY,
		COMEDI_AD_ANALOG);
	addparam(&dev->chinfo[16].paramlist,COMEDI_IO,COMEDI_RDONLY,
		COMEDI_IO_OUTPUT);
	addparam(&dev->chinfo[16].paramlist,COMEDI_LASTDATA,
		COMEDI_RDONLY,0);
	dev->chinfo[16].lastdata=dev->chinfo[i].paramlist;

	dev->chinfo[17].trig=dt2811_da;
	dev->chinfo[17].sp=NULL;
	addparam(&dev->chinfo[17].paramlist,COMEDI_BITS,COMEDI_RDONLY,
		12);
	addparam(&dev->chinfo[17].paramlist,COMEDI_MAXDATA,COMEDI_RDONLY,
		0xfff);
	addparam(&dev->chinfo[17].paramlist,COMEDI_AD,COMEDI_RDONLY,
		COMEDI_AD_ANALOG);
	addparam(&dev->chinfo[17].paramlist,COMEDI_IO,COMEDI_RDONLY,
		COMEDI_IO_OUTPUT);
	addparam(&dev->chinfo[17].paramlist,COMEDI_LASTDATA,
		COMEDI_RDONLY,0);
	dev->chinfo[17].lastdata=dev->chinfo[i].paramlist;

	dev->chinfo[18].trig=dt2811_di;
	dev->chinfo[18].sp=NULL;
	addparam(&dev->chinfo[18].paramlist,COMEDI_BITS,COMEDI_RDONLY,
		8);
	addparam(&dev->chinfo[18].paramlist,COMEDI_MAXDATA,COMEDI_RDONLY,
		0xff);
	addparam(&dev->chinfo[18].paramlist,COMEDI_AD,COMEDI_RDONLY,
		COMEDI_AD_DIGITAL);
	addparam(&dev->chinfo[18].paramlist,COMEDI_IO,COMEDI_RDONLY,
		COMEDI_IO_INPUT);
	addparam(&dev->chinfo[18].paramlist,COMEDI_LASTDATA,
		COMEDI_RDONLY,0);
	dev->chinfo[18].lastdata=dev->chinfo[i].paramlist;

	dev->chinfo[19].trig=dt2811_do;
	dev->chinfo[19].sp=NULL;
	addparam(&dev->chinfo[19].paramlist,COMEDI_BITS,COMEDI_RDONLY,8);
	addparam(&dev->chinfo[19].paramlist,COMEDI_MAXDATA,COMEDI_RDONLY,0xff);
	addparam(&dev->chinfo[19].paramlist,COMEDI_AD,COMEDI_RDONLY,
		COMEDI_AD_DIGITAL);
	addparam(&dev->chinfo[19].paramlist,COMEDI_IO,COMEDI_RDONLY,
		COMEDI_IO_INPUT);
	addparam(&dev->chinfo[19].paramlist,COMEDI_LASTDATA,
		COMEDI_RDONLY,0);
	dev->chinfo[19].lastdata=dev->chinfo[i].paramlist;

	comedi_initbuf(&dev->rdbuf);
	
	dev->rem=dt2811_rem;

	return 1;
}


static int dt2811_rem(comedi_device *dev)
{
	int i;

	printk("comedi%d: dt2811: remove\n",dev->minor);
	
	for(i=0;i<dev->chan;i++)
		delparamlist(dev->chinfo[i].paramlist);
	kfree(dev->chinfo);

	if(dev->irq){
		free_irq(dev->irq,dev);
	}
	release_region(dev->iobase,dev->iosize);

	comedi_freebuf(&dev->rdbuf);
	
	return 0;
}

static int dt2811_ad(comedi_device *dev,comedi_trig *it)
{
	int lo,hi;
	
	outb(it->chan,dev->iobase+DT2811_ADGCR);
	while(inb(dev->iobase+DT2811_ADCSR)&DT2811_ADBUSY)
		udelay(25);
	lo=inb(dev->iobase+DT2811_ADDATLO);
	hi=inb(dev->iobase+DT2811_ADDATHI);

	it->data=lo+0x100*hi;
	
	return 0;
}

#if 0
int dt2811_adtrig(kdev_t minor,comedi_adtrig *adtrig)
{
	comedi_device *dev=comedi_devices+minor;

	if(adtrig->n<1)return 0;
	dev->curadchan=adtrig->chan;
switch(dev->i_admode){
case COMEDI_MDEMAND:
	dev->ntrig=adtrig->n-1;
	/*printk("dt2811: AD soft trigger\n");*/
	/*outb(DT2811_CLRERROR|DT2811_INTENB,dev->iobase+DT2811_ADCSR);*/  /* not neccessary */
	outb(dev->curadchan,dev->iobase+DT2811_ADGCR);
	do_gettimeofday(&trigtime);
	break;
case COMEDI_MCONTS:
	dev->ntrig=adtrig->n;
	break;
}

	return 0;
}
#endif

static int dt2811_da(comedi_device *dev,comedi_trig *it)
{
	int lo,hi;

	lo=(it->data&0xff);
	hi=(it->data>>8)&0x0f;
	
	switch(it->chan){
	case 16:
		outb(lo,dev->iobase+DT2811_DADAT0LO);
		outb(hi,dev->iobase+DT2811_DADAT0HI);
		break;
	case 17:
		outb(lo,dev->iobase+DT2811_DADAT1LO);
		outb(hi,dev->iobase+DT2811_DADAT1HI);
		break;
	}
	return 0;
}

static int dt2811_di(comedi_device *dev,comedi_trig *it)
{
	comedi_sample s;
	
	s.data=it->data=inb(dev->iobase+DT2811_DIO);
	s.chan=it->chan;
	
	return 0;
}

static int dt2811_do(comedi_device *dev,comedi_trig *it)
{
	outb(it->data,dev->iobase+DT2811_DIO);

	return 0;
}

