/*
    module/dt2814.c
    hardware driver for Data Translation DT2814

    COMEDI - Linux Control and Measurement Device Interface
    Copyright (C) 1998 David A. Schleef <ds@stm.lbl.gov>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/


#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/mm.h>
#include <linux/malloc.h>
#include <linux/errno.h>
#include <linux/ioport.h>
#include <linux/delay.h>
#include <linux/interrupt.h>
#include <linux/timex.h>
#include <linux/timer.h>
#include <asm/io.h>
#include <comedi_module.h>



#define DT2814_SIZE 2

#define DT2814_CSR 0
#define DT2814_DATA 1

/*
 * flags
 */

#define DT2814_FINISH 0x80
#define DT2814_ERR 0x40
#define DT2814_BUSY 0x20
#define DT2814_ENB 0x10
#define DT2814_CHANMASK 0x0f

comedi_devinit dt2814_init;
static int dt2814_ad(comedi_device *dev,comedi_trig *it);
static int dt2814_rem(comedi_device *dev);
static void dt2814_interrupt(int irq,void *dev,struct pt_regs * regs);

typedef struct{
	int ntrig;
	int curadchan;
}dt2814_private;
#define devpriv ((dt2814_private *)dev->private)


#define DT2814_TIMEOUT 100

static int dt2814_iad(comedi_device *dev,comedi_trig *it)
{
	int i,hi,lo;
	unsigned long flags;

	if(dev->busy[0])return -EBUSY;

	outb(it->chan,dev->iobase+DT2814_CSR);
	save_flags(flags);
	cli();  /* FIXME */
	for(i=0;i<DT2814_TIMEOUT;i++){
		if(inb(dev->iobase+DT2814_CSR)&DT2814_FINISH)
			break;
	}
	hi=inb(dev->iobase+DT2814_DATA);
	lo=inb(dev->iobase+DT2814_DATA);
	restore_flags(flags);

	it->data=(hi<<4)|(lo>>4);
	
	return 0;
}

static int dt2814_ad(comedi_device *dev,comedi_trig *it)
{
	if(dev->busy[0])return -EBUSY;

	switch(it->trigsrc){
	case COMEDI_TRIGCLK:
		devpriv->ntrig=it->n;
		outb(it->chan|DT2814_ENB|(it->trigvar<<5),
			dev->iobase+DT2814_CSR);
		dev->busy[0]=it->job;
		break;
	default:
		return -EINVAL;
	}
	
	return 0;
}

static char *driver_name="dt2814";

int dt2814_init(comedi_device *dev,comedi_devconfig *it)
{
	int i,irqs,irq;
	long flags;

	if(strcmp(driver_name,it->name))
		return 0;

	dev->iobase=it->options[0];
	printk("comedi%d: dt2814: 0x%04x ",dev->minor,dev->iobase);
	if(check_region(dev->iobase,DT2814_SIZE)<0){
		printk("I/O port conflict\n");
		return -EIO;
	}
	request_region(dev->iobase,DT2814_SIZE,driver_name);
	dev->board_name=driver_name;
	dev->driver_name=driver_name;
	dev->iobase=dev->iobase;
	dev->iosize=DT2814_SIZE;

	outb(0,dev->iobase+DT2814_CSR);
	udelay(100);
	if(inb(dev->iobase+DT2814_CSR)&DT2814_ERR){
		printk("reset error (fatal)\n");
		return -EIO;
	}
	i=inb(dev->iobase+DT2814_DATA);
	i=inb(dev->iobase+DT2814_DATA);

	irq=it->options[1];
	if(irq<0){
		save_flags(flags);
		sti();
		irqs=probe_irq_on();
	
		outb(0,dev->iobase+DT2814_CSR);

		udelay(100);

		irq=probe_irq_off(irqs);
		restore_flags(flags);
		if(inb(dev->iobase+DT2814_CSR)&DT2814_ERR){
			printk("error probing irq (bad) \n");
		}

		i=inb(dev->iobase+DT2814_DATA);
		i=inb(dev->iobase+DT2814_DATA);
	}
	dev->irq=0;
	if(irq>0){
		printk("( irq = %d )\n",irq);
		request_irq(irq,dt2814_interrupt,0*SA_INTERRUPT,"dt2814",dev);
		dev->irq=irq;
	}else if(irq==0){
		printk("(no irq)\n");
	}else{
		printk("(probe returned multiple irqs--bad)\n");
	}
	
	dev->chan=16;

	dev->private=kmalloc(sizeof(dt2814_private)+
		dev->chan*sizeof(comedi_chinfo),GFP_KERNEL);
	if(!dev->private){
		return -ENOMEM;
	}
	memset(dev->private,0,sizeof(dt2814_private)+
		dev->chan*sizeof(comedi_chinfo));
	dev->chinfo=dev->private+sizeof(dt2814_private);
	for(i=0;i<dev->chan;i++){
		dev->chinfo[i].trig=dt2814_ad;
		dev->chinfo[i].itrig=dt2814_iad;
		dev->chinfo[i].sp=NULL;
		addparam(&dev->chinfo[i].paramlist,COMEDI_BITS,COMEDI_RDONLY,
			12);
		addparam(&dev->chinfo[i].paramlist,COMEDI_MAXDATA,COMEDI_RDONLY,
			0xfff);
		addparam(&dev->chinfo[i].paramlist,COMEDI_AD,COMEDI_RDONLY,
			COMEDI_AD_ANALOG);
		addparam(&dev->chinfo[i].paramlist,COMEDI_IO,COMEDI_RDONLY,
			COMEDI_IO_INPUT);
		addparam(&dev->chinfo[i].paramlist,COMEDI_LASTDATA,
			COMEDI_RDONLY,0);
		dev->chinfo[i].lastdata=dev->chinfo[i].paramlist;
	}

	comedi_initbuf(&dev->rdbuf);

	dev->rem=dt2814_rem;

	return 1;
}


static int dt2814_rem(comedi_device *dev)
{
	int i;
	
	printk("comedi%d: dt2814: remove\n",dev->minor);
	
	for(i=0;i<dev->chan;i++)
		delparamlist(dev->chinfo[i].paramlist);
	kfree(dev->private);

	if(dev->irq){
		free_irq(dev->irq,dev);
	}
	release_region(dev->iobase,dev->iosize);

	comedi_freebuf(&dev->rdbuf);

	return 0;
}


static void dt2814_interrupt(int irq,void *d,struct pt_regs * regs)
{
        int lo,hi;
        comedi_device *dev=d;
        comedi_sample pt;

        hi=inb(dev->iobase+DT2814_DATA);
        lo=inb(dev->iobase+DT2814_DATA);

        pt.data=(hi<<4)|(lo>>4);
        pt.chan=devpriv->curadchan;
	report_sample(dev,&pt);

	if(!(--devpriv->ntrig)){
		int flags,i;

		outb(0,dev->iobase+DT2814_CSR);
		/* note: turning off timed mode triggers another
			sample. */

		save_flags(flags);
		cli();	/* FIXME */
		for(i=0;i<DT2814_TIMEOUT;i++){
			if(inb(dev->iobase+DT2814_CSR)&DT2814_FINISH)
				break;
		}
		inb(dev->iobase+DT2814_DATA);
		inb(dev->iobase+DT2814_DATA);
		restore_flags(flags);
	
		dev->busy[0]=0;
		wake_up_interruptible(&dev->rdwait);
	}
}


