/*
    module/module.c
    comedi kernel module

    COMEDI - Linux Control and Measurement Device Interface
    Copyright (C) 1997-8 David A. Schleef <ds@stm.lbl.gov>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#if !OLD_KERNEL
#define EXPORT_SYMTAB
#endif

#include <linux/module.h>

#include <linux/errno.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/fcntl.h>
#include <linux/delay.h>
#include <linux/ioport.h>
#include <linux/mm.h>
#include <linux/malloc.h>
#include <asm/io.h>
#include <comedi_module.h>

comedi_device *comedi_devices;
unsigned int comedi_job=1;

static int comedi_ioctl(struct inode * inode,struct file * file,unsigned int cmd,unsigned long arg)
{
	int retval,i;
	kdev_t minor=MINOR(inode->i_rdev);
	comedi_device *dev=comedi_devices+minor;

	switch(cmd){
	case COMEDI_CONFIG: /* for overall configuation */
	{
		comedi_devconfig db;

#ifdef DEBUG
		printk("entering COMEDI_CONFIG\n");
#endif
		if(!suser())
			return -EPERM;
#if OLD_KERNEL
		retval=verify_area(VERIFY_READ,(void *)arg,
			sizeof(comedi_devconfig));
		if(retval)
			return retval;
		memcpy_fromfs(&db,(void *)arg,sizeof(comedi_devconfig));
#else
		memcpy(&db,(void *)arg,sizeof(comedi_devconfig));
#endif
#if DEBUG
		printk("memcpy ok\n");
#endif
		if(dev->rem){
#if DEBUG
			printk("trying to remove existing device %08x\n",
				dev->rem);
#endif
			retval=dev->rem(dev);
			if(retval<0)return retval;
		}
		memset(dev,0,sizeof(comedi_device));
#if DEBUG
		printk("memset ok\n");
#endif
		dev->minor=minor;
		if(!db.name[0])return 0;
#if DEBUG
		printk("scanning devinits\n");
#endif
		for(i=0;i<comedi_maxinits;i++){
			retval=comedi_devinits[i](dev,&db);
			if(retval<0)
				return retval;
			if(retval)return 0;
		}
#if DEBUG
		printk("scanning devinits failed\n");
#endif
		return -EINVAL;
	}
	case COMEDI_GETINFO: /* returns overall info */
	{
		comedi_info it;
		
		memset(&it,0,sizeof(comedi_info));
		
		if(dev->board_name)
			strncpy(it.board_name,dev->board_name,COMEDI_NAMELEN-1);
		if(dev->driver_name)
			strncpy(it.driver_name,dev->driver_name,COMEDI_NAMELEN-1);
		it.chan=dev->chan;
		
		it.iobase=dev->iobase;
		it.iosize=dev->iosize;
		it.irq=dev->irq;
		it.minor=dev->minor;
		it.version_code=COMEDI_VERSION_CODE;
		
#if OLD_KERNEL
		retval=verify_area(VERIFY_WRITE,(void *)arg,sizeof(comedi_info));
		if(retval)
			return retval;
		memcpy_tofs((void *)arg,&it,sizeof(comedi_info));
#else
		memcpy((void *)arg,&it,sizeof(comedi_info));
#endif
		return 0;
	}
	case COMEDI_GETPARAM:
	{
		comedi_param param;
		comedi_iparam *p;
		
#if OLD_KERNEL
		retval=verify_area(VERIFY_READ,(void *)arg,sizeof(comedi_param));
		if(retval)return retval;
		memcpy_fromfs(&param,(void *)arg,sizeof(comedi_param));
#else
		memcpy(&param,(void *)arg,sizeof(comedi_param));
#endif
		p=NULL;
		if(param.chan<dev->chan)p=dev->chinfo[param.chan].paramlist;
		while(p){
			if(p->pnum==param.pnum){
				param.flags=p->flags;
				param.pval=p->pval;
#if OLD_KERNEL
				retval=verify_area(VERIFY_WRITE,(void *)arg,
					sizeof(comedi_param));
				if(retval)return retval;
				memcpy_tofs((void *)arg,&param,
					sizeof(comedi_param));
#else
				memcpy((void *)arg,&param,sizeof(comedi_param));
#endif
				return 0;
			}else{
				p=p->next;
			}
		}
		return -EINVAL;
	}
	case COMEDI_SETPARAM:
	{
		comedi_param param;
		
#if OLD_KERNEL
		retval=verify_area(VERIFY_READ,(void *)arg,sizeof(comedi_param));
		if(retval)return retval;
		memcpy_fromfs(&param,(void *)arg,sizeof(comedi_param));
#else
		memcpy(&param,(void *)arg,sizeof(comedi_param));
#endif
#ifdef CONFIG_LOCK
		if(dev->chinfo[param.chan].lock &&
		   dev->chinfo[param.chan].lock!=file)
			return -EACCES;
#endif
		if(param.chan<dev->chan && dev->chinfo[param.chan].sp)
			return dev->chinfo[param.chan].sp(dev,&param);
		return -EINVAL;
	}
	case COMEDI_GETNPARAM:
	{
		comedi_param param;
		comedi_iparam *p;
		
#if OLD_KERNEL
		retval=verify_area(VERIFY_READ,(void *)arg,sizeof(comedi_param));
		if(retval)return retval;
		memcpy_fromfs(&param,(void *)arg,sizeof(comedi_param));
#else
		memcpy(&param,(void *)arg,sizeof(comedi_param));
#endif
		p=NULL;
		if(param.chan<dev->chan)p=dev->chinfo[param.chan].paramlist;
		if(!p)return -EINVAL;
		while(param.pnum--){
			p=p->next;
			if(!p)return -EINVAL;
		}
		param.pnum=p->pnum;
		param.flags=p->flags;
		param.pval=p->pval;
#if OLD_KERNEL
		retval=verify_area(VERIFY_WRITE,(void *)arg,
			sizeof(comedi_param));
		if(retval)return retval;
		memcpy_tofs((void *)arg,&param,
			sizeof(comedi_param));
#else
		memcpy((void *)arg,&param,sizeof(comedi_param));
#endif
		return 0;
	}
	case COMEDI_TRIG: /* triggers conversions */
	{
		comedi_trig it;
		
#if OLD_KERNEL
		retval=verify_area(VERIFY_READ,(void *)arg,sizeof(comedi_trig));
		if(retval)return retval;
		retval=verify_area(VERIFY_WRITE,(void *)arg,sizeof(comedi_trig));
		if(retval)return retval;
		memcpy_fromfs(&it,(void *)arg,sizeof(comedi_trig));
#else
		memcpy(&it,(void *)arg,sizeof(comedi_trig));
#endif
		if(it.chan>=dev->chan)return -EINVAL;
#ifdef CONFIG_LOCK
		if(dev->chinfo[it.chan].lock &&
		   dev->chinfo[it.chan].lock!=file)
			return -EACCES;
#endif
		if(it.trigsrc==COMEDI_TRIGNOW){
			if(!dev->chinfo[it.chan].itrig)
				return -EINVAL;
			it.job=comedi_job++;
			retval=dev->chinfo[it.chan].itrig(dev,&it);
		}else{
			if(!dev->chinfo[it.chan].trig)
				return -EINVAL;
			it.job=comedi_job++;
			retval=dev->chinfo[it.chan].trig(dev,&it);
		}
#if OLD_KERNEL
		memcpy_tofs((void *)arg,&it,sizeof(comedi_trig));
#else
		memcpy((void *)arg,&it,sizeof(comedi_trig));
#endif
		return retval;
	}
	case COMEDI_MTRIG: /* triggers multi-channel conversions */
	{
		comedi_mtrig it;
		unsigned int *chanlist;
		
#if OLD_KERNEL
		retval=verify_area(VERIFY_READ,(void *)arg,sizeof(comedi_mtrig));
		if(retval)return retval;
		retval=verify_area(VERIFY_WRITE,(void *)arg,sizeof(comedi_mtrig));
		if(retval)return retval;
		memcpy_fromfs(&it,(void *)arg,sizeof(comedi_mtrig));
#else
		memcpy(&it,(void *)arg,sizeof(comedi_mtrig));
#endif
		if(it.n1>COMEDI_MAX_MCHAN){
			return -EINVAL;
		}
		if(!dev->mtrig){
			return -EINVAL;
		}
#if OLD_KERNEL
		retval=verify_area(VERIFY_READ,it.chanlist,it.n1*sizeof(int));
		if(retval)return retval;
		
		chanlist=kmalloc(it.n1*sizeof(int),GFP_KERNEL);
		if(!chanlist)return -ENOMEM;
		memcpy_fromfs(chanlist,it.chanlist,it.n1*sizeof(int));
#else
		chanlist=kmalloc(it.n1*sizeof(int),GFP_KERNEL);
		if(!chanlist)return -ENOMEM;
		memcpy(chanlist,it.chanlist,it.n1*sizeof(int));
#endif
		
		it.job=comedi_job++;
		retval=dev->mtrig(dev,&it,chanlist);
#if OLD_KERNEL
		memcpy_tofs((void *)arg,&it,sizeof(comedi_mtrig));
#else
		memcpy((void *)arg,&it,sizeof(comedi_mtrig));
#endif
		return retval;
	}
#if OLD_KERNEL
#if 0
	case COMEDI_COPY_TO:
	{
		comedi_memcpy it;

		if(!dev->memcpy_to)
			return -EIO;

		retval=verify_area(VERIFY_READ,(void *)arg,sizeof(comedi_memcpy));
		if(retval)return retval;
		memcpy_fromfs(&it,(void *)arg,sizeof(comedi_memcpy));
		
		retval=verify_area(VERIFY_READ,it.local,it.nbytes);
		if(retval)return retval;

		retval=dev->memcpy_to(dev,&it);
		return retval;
	}
#endif
	case COMEDI_COPY_FROM:
	{
		comedi_memcpy it;
		comedi_memcpy it_fake;

		if(!dev->memcpy_from){
			return -EIO;
		}
	
		retval=verify_area(VERIFY_READ,(void *)arg,sizeof(comedi_memcpy));
		if(retval)return retval;
		memcpy_fromfs(&it,(void *)arg,sizeof(comedi_memcpy));
		
		retval=verify_area(VERIFY_WRITE,it.local,it.nbytes);
		if(retval)return retval;
		
		it_fake=it;
		it_fake.local=kmalloc(it.nbytes,GFP_KERNEL);
		if(!it_fake.local)return -ENOMEM;

		retval=dev->memcpy_from(dev,&it_fake);
		
		memcpy_tofs(it.local,it_fake.local,it.nbytes);
		kfree(it_fake.local);
		
		return retval;
	}
#else
	case COMEDI_COPY_TO:
	{
		comedi_memcpy *it;

		if(!dev->memcpy_to)
			return -EIO;
		it=(void *)arg;

		retval=dev->memcpy_to(dev,it);
		return retval;
	}
	case COMEDI_COPY_FROM:
	{
		comedi_memcpy *it;

		if(!dev->memcpy_from){
			return -EIO;
		}
		it=(void *)arg;
	
		retval=dev->memcpy_from(dev,it);
		return retval;
	}
	case COMEDI_MEMSET:
	{
		comedi_memcpy *it;

		if(!dev->d_memset)
			return -EIO;
		it=(void *)arg;
	
		return dev->d_memset(dev,it);
	}
	case COMEDI_EXEC:
		if(!dev->memexec)
			return -EIO;
		return dev->memexec(dev,arg);
	case COMEDI_RESET:
		if(!dev->reset)
			return -EIO;
		return dev->reset(dev,arg);
#endif
#ifdef CONFIG_LOCK
	case COMEDI_LOCK:
	{
		unsigned int chan=arg;
	
printk("comedi: locking chan %d key %p\n",chan,file);
		if(chan>=dev->chan)return -EINVAL;
		if(dev->chinfo[chan].lock &&
		   dev->chinfo[chan].lock!=file)
			return -EACCES;

		dev->chinfo[chan].lock=file;
		
		return 0;
	}
	case COMEDI_UNLOCK:
	{
		unsigned int chan=arg;
	
printk("comedi: unlocking chan %d key %p\n",chan,file);
		if(chan>=dev->chan)return -EINVAL;
		if(dev->chinfo[chan].lock &&
		   dev->chinfo[chan].lock!=file)
			return -EACCES;
		dev->chinfo[chan].lock=NULL;

		return 0;
	}
#endif
	default:
		return -EIO;
	}
}

#if OLD_KERNEL
static int comedi_write(struct inode *inode,struct file *file,const char *buf,int nbytes)
#else
static ssize_t comedi_write(struct file *file,const char *buf,size_t nbytes,loff_t *offset)
#endif
{
#if OLD_KERNEL
	kdev_t minor=MINOR(inode->i_rdev);
	int retval;
#else
	kdev_t minor=MINOR(file->f_dentry->d_inode->i_rdev);
#endif
	comedi_device *dev=comedi_devices+minor;
	/*struct wait_queue wait={current,NULL};*/

	if(!dev->write)return -EIO;
	if(!nbytes)return 0;

#if OLD_KERNEL
	retval=verify_area(VERIFY_READ,(void *)buf,nbytes);
	if(retval)
		return retval;
#endif
	return dev->write(dev,buf,nbytes);
}

#if OLD_KERNEL
static int comedi_read(struct inode *inode,struct file * file,char *buf,int nbytes)
#else
static ssize_t comedi_read(struct file * file,char *buf,size_t nbytes,loff_t *offset)
#endif
{
#if OLD_KERNEL
	kdev_t minor=MINOR(inode->i_rdev);
#else
	kdev_t minor=MINOR(file->f_dentry->d_inode->i_rdev);
#endif
	comedi_device *dev=comedi_devices+minor;
	int n,count=0,retval=0;
	struct wait_queue wait={current,NULL};

	if(!file->private_data)return -EIO;
	if(!nbytes)return 0;

	add_wait_queue(&dev->rdwait,&wait);
	current->state=TASK_INTERRUPTIBLE;
	while(nbytes>0){
		n=copy_from_buffer((comedi_bufptr *)(file->private_data),buf+count,nbytes);
		if(!n){
			if(file->f_flags&O_NONBLOCK){
				retval=-EAGAIN;
				break;
			}
#if OLD_KERNEL
			if(current->signal & ~current->blocked){
#else
			if(signal_pending(current)){
#endif
				retval=-ERESTARTSYS;
				break;
			}
			schedule();
			continue;
		}
		count+=n;
		nbytes-=n;
	}
	current->state=TASK_RUNNING;
	remove_wait_queue(&dev->rdwait,&wait);

	return (count ? count : retval);
}

/*
static int comedi_select(struct inode * inode,struct file * file,int sel_type,select_table *wait)
{
	return 0;
}
*/

#if OLD_KERNEL
static int comedi_lseek(struct inode * inode,struct file *file,off_t offset,int origin)
#else
static loff_t comedi_lseek(struct file *file,loff_t offset,int origin)
#endif
{
	return -ESPIPE;
}

static int comedi_open(struct inode *inode,struct file *file)
{
	kdev_t minor=MINOR(inode->i_rdev);
	comedi_device *dev=comedi_devices+minor;

	if(minor>=COMEDI_NDEVICES)return -ENODEV;
	if(!comedi_devices[minor].rem && !suser())
		return -ENODEV;
	MOD_INC_USE_COUNT;
	comedi_devices[minor].use_count++;
	
	if(dev->rdbuf.last)
		file->private_data=(void *)comedi_hookbuf(&(dev->rdbuf));
	else file->private_data=NULL;

	return 0;
}

#if OLD_KERNEL
static void comedi_close(struct inode *inode,struct file *file)
#else
static int comedi_close(struct inode *inode,struct file *file)
#endif
{
	kdev_t minor=MINOR(inode->i_rdev);
	comedi_device *dev=comedi_devices+minor;
#ifdef CONFIG_LOCK
	int i;
#endif

#ifdef CONFIG_LOCK
	for(i=0;i<dev->chan;i++){
		if(dev->chinfo[i].lock==file)
			dev->chinfo[i].lock=NULL;
	}
#endif
	if(file->private_data)
		comedi_unhookbuf((comedi_bufptr *)(file->private_data));
	MOD_DEC_USE_COUNT;
	dev->use_count--;
	
#if !OLD_KERNEL
	return 0;
#endif
}

static struct file_operations comedi_fops={
	comedi_lseek,
	comedi_read,
	comedi_write,
	NULL, /* comedi_readdir */
	NULL, /* comedi_poll, */
	comedi_ioctl,
	NULL, /* comedi_mmap */
	comedi_open,
#if !OLD_KERNEL
	NULL,		/* flush */
#endif
	comedi_close,
	NULL, /* fsync */
	NULL, /* fasync */
	NULL, /* check_media_change */
	NULL /* revalidate */
#if !OLD_KERNEL
	,
	NULL	/* lock */

#endif
};

extern struct symbol_table comedi_syms;

int init_module(void)
{
	printk("comedi: version " COMEDI_VERSION " - David Schleef <ds@stm.lbl.gov>\n");
	if(register_chrdev(COMEDI_MAJOR,"comedi",&comedi_fops)){
		printk("comedi: unable to get major %d\n",COMEDI_MAJOR);
		return -EIO;
	}
	comedi_devices=(comedi_device *)kmalloc(sizeof(comedi_device)*COMEDI_NDEVICES,GFP_KERNEL);
	if(!comedi_devices)
		return -ENOMEM;
	memset(comedi_devices,0,sizeof(comedi_device)*COMEDI_NDEVICES);
	init_param();
	init_polling();
#if OLD_KERNEL
	register_symtab(&comedi_syms);
#endif
	
	return 0;
}

void cleanup_module(void)
{
	int i;

	if(MOD_IN_USE)
		printk("comedi: module in use -- remove delayed\n");
	for(i=0;i<COMEDI_NDEVICES;i++){
		if(comedi_devices[i].rem)
			comedi_devices[i].rem(comedi_devices+i);
	}
	cleanup_polling();
	cleanup_param();
	kfree(comedi_devices);

	unregister_chrdev(COMEDI_MAJOR,"comedi");
}

void report_sample(comedi_device *dev,comedi_sample *s)
{
	copy_to_buffer(&(dev->rdbuf),s,sizeof(comedi_sample));
	if(dev->chinfo[s->chan].lastdata)
		dev->chinfo[s->chan].lastdata->pval=s->data;
}

void comedi_error(comedi_device *dev,const char *s)
{
	printk("comedi%d: %s: %s\n",dev->minor,dev->driver_name,s);
}



/*
	start of exported functions.  compilation chokes if this
	is in a separate file.
*/


int comedi_get(int minor,int chan,sampl_t *data);
int comedi_put(int minor,int chan,sampl_t *data);



int comedi_get(int minor,int chan,sampl_t *data)
{
	comedi_device *dev;
	comedi_trig t;
	int ret;
	
	if(minor<0 || minor>=COMEDI_NDEVICES)return -ENODEV;
	dev=comedi_devices+minor;
	if(!dev->rem)return -EINVAL;
	if(!dev->chinfo[chan].itrig)return -EINVAL;
	
	if(chan<0 || chan>=dev->chan)return -EINVAL;
	t.chan=chan;
	t.io=COMEDI_IO_INPUT;
	t.n=1;
	t.trigsrc=COMEDI_TRIGNOW;
	
	ret=dev->chinfo[chan].itrig(dev,&t);
	
	*data=t.data;
	
	return ret;
}

int comedi_put(int minor,int chan,sampl_t *data)
{
	comedi_device *dev;
	comedi_trig t;
	int ret;

	if(minor<0 || minor>=COMEDI_NDEVICES)return -ENODEV;
	dev=comedi_devices+minor;
	if(!dev->rem)return -EINVAL;
	if(!dev->chinfo[chan].itrig)return -EINVAL;
	
	if(chan<0 || chan>=dev->chan)return -EINVAL;
	t.chan=chan;
	t.data=*data;
	t.io=COMEDI_IO_OUTPUT;
	t.n=1;
	t.trigsrc=COMEDI_TRIGNOW;
	
	ret=dev->chinfo[chan].itrig(dev,&t);
	
	return ret;
}

#if OLD_KERNEL

static struct symbol_table comedi_syms = {
#include <linux/symtab_begin.h>
	X(comedi_get),
	X(comedi_put),
#include <linux/symtab_end.h>
};

#else

EXPORT_SYMBOL(comedi_get);
EXPORT_SYMBOL(comedi_put);

#endif

