/*
    module/parport.c
    hardware driver for standard parallel port

    COMEDI - Linux Control and Measurement Device Interface
    Copyright (C) 1998 David A. Schleef <ds@stm.lbl.gov>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

/*

   TODO:

   - support bit mask ioctl
   - EPP/ECP support

 */

#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/mm.h>
#include <linux/malloc.h>
#include <linux/errno.h>
#include <linux/ioport.h>
#include <linux/delay.h>
#include <linux/interrupt.h>
#include <linux/timex.h>
#include <linux/timer.h>
#include <asm/io.h>
#include <comedi_module.h>


#define PARPORT_SIZE 3

#define PARPORT_A 0
#define PARPORT_B 1
#define PARPORT_C 2

typedef struct{
	int cr;
}parport_private;
#define devpriv ((parport_private *)dev->private)

comedi_devinit parport_init;
static int parport_rem(comedi_device *dev);

static int parport_dio_a(comedi_device *dev,comedi_trig *it)
{
	switch(it->io){
	case COMEDI_IO_INPUT:
		break;
	case COMEDI_IO_OUTPUT:
		outb(it->data,dev->iobase+PARPORT_A);
		return 0;
	}
	return -EINVAL;
}

static int parport_dio_b(comedi_device *dev,comedi_trig *it)
{
	switch(it->io){
	case COMEDI_IO_INPUT:
		it->data=inb(dev->iobase+PARPORT_B);
		it->data>>=3;
		return 0;
	case COMEDI_IO_OUTPUT:
		break;
	}
	return -EINVAL;
}

static int parport_dio_c(comedi_device *dev,comedi_trig *it)
{
	switch(it->io){
	case COMEDI_IO_INPUT:
		break;
	case COMEDI_IO_OUTPUT:
		outb(it->data & 0xf,dev->iobase+PARPORT_C);
		return 0;
	}
	return -EINVAL;
}


int parport_init(comedi_device *dev,comedi_devconfig *it)
{
	int i;

	if(strcmp("parport",it->name))
		return 0;

	dev->iobase=it->options[0];
	printk("comedi%d: parport: 0x%04x ",dev->minor,dev->iobase);
	if(check_region(dev->iobase,PARPORT_SIZE)<0){
		printk("I/O port conflict\n");
		return -EIO;
	}
	request_region(dev->iobase,PARPORT_SIZE,"parport (comedi)");
	dev->iosize=PARPORT_SIZE;
	dev->irq=0;
	dev->board_name="parport";
	dev->driver_name="parport";

	dev->chan=3;

	devpriv=kmalloc(sizeof(parport_private)+dev->chan*sizeof(comedi_chinfo),
		GFP_KERNEL);
	if(!devpriv)return -ENOMEM;
	memset(devpriv,0,sizeof(parport_private)+dev->chan*sizeof(comedi_chinfo));

	dev->chinfo=(void *)(devpriv+1);

	i=0;
	dev->chinfo[i].trig=parport_dio_a;
	addparam(&dev->chinfo[i].paramlist,COMEDI_BITS,COMEDI_RDONLY,8);
	addparam(&dev->chinfo[i].paramlist,
		COMEDI_AD,COMEDI_RDONLY,COMEDI_AD_DIGITAL);
	addparam(&dev->chinfo[i].paramlist,
		COMEDI_IO,COMEDI_RDONLY,COMEDI_IO_OUTPUT);
	
	i++;
	dev->chinfo[i].trig=parport_dio_b;
	addparam(&dev->chinfo[i].paramlist,COMEDI_BITS,COMEDI_RDONLY,4);
	addparam(&dev->chinfo[i].paramlist,
		COMEDI_AD,COMEDI_RDONLY,COMEDI_AD_DIGITAL);
	addparam(&dev->chinfo[i].paramlist,
		COMEDI_IO,COMEDI_RDONLY,COMEDI_IO_INPUT);
	
	i++;
	dev->chinfo[i].trig=parport_dio_c;
	addparam(&dev->chinfo[i].paramlist,COMEDI_BITS,COMEDI_RDONLY,4);
	addparam(&dev->chinfo[i].paramlist,
		COMEDI_AD,COMEDI_RDONLY,COMEDI_AD_DIGITAL);
	addparam(&dev->chinfo[i].paramlist,
		COMEDI_IO,COMEDI_RDONLY,COMEDI_IO_OUTPUT);

	comedi_initbuf(&dev->rdbuf);

	dev->rem=parport_rem;

	printk("\n");
	return 1;
}


static int parport_rem(comedi_device *dev)
{
	int i;

	printk("comedi%d: parport: remove\n",dev->minor);
	
	for(i=0;i<dev->chan;i++)
		delparamlist(dev->chinfo[i].paramlist);
	kfree(devpriv);
	release_region(dev->iobase,dev->iosize);

	comedi_freebuf(&dev->rdbuf);

	return 0;
}

