/*
    module/pci-mio.c
    Hardware driver for NI PCI-MIO E series cards

    COMEDI - Linux Control and Measurement Device Interface
    Copyright (C) 1997-8 David A. Schleef <ds@stm.lbl.gov>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

/*
	Interrupt support added by Truxton Fulton <trux@truxton.com>

	References for specifications:
	
	   321747b.pdf  Register Level Programmer Manual (obsolete)
	   321747c.pdf  Register Level Programmer Manual (new)
	   DAQ-STC reference manual

	Other possibly relevant info:
	
	   320517c.pdf  User manual (obsolete)
	   320517f.pdf  User manual (new)
	   320889a.pdf  delete
	   320906c.pdf  maximum signal ratings
	   321066a.pdf  about 16x
	   321791a.pdf  discontinuation of at-mio-16e-10 rev. c
	   321808a.pdf  about at-mio-16e-10 rev P
	   321837a.pdf  discontinuation of at-mio-16de-10 rev d
	   321838a.pdf  about at-mio-16de-10 rev N
	
	ISSUES:

	need to deal with external reference for DAC, and other DAC
	properties in board properties
	
	deal with at-mio-16de-10 revision D to N changes, etc.
	
*/

#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/ioport.h>
#include <linux/delay.h>
#include <linux/mm.h>
#include <linux/interrupt.h>
#include <asm/io.h>
#include <linux/malloc.h>
#include <comedi_module.h>
#include <ni_stc.h>

#undef DEBUG

#define NI_SIZE 0x20


static int ni_modebits1[4]={ 0x3000, 0x2000, 0x1000, 0 };
static int ni_modebits2[4]={ 0x3f, 0x3f, 0x37, 0x37 };

struct ni_boardtype_struct{
	int boardcode;
	char *name;
	int n_adchan;
	int adbits;
	int alwaysdither;
	int limitedgain;
	int caldactype;
	int eepromlayout;
	int n_ao;
	int can_do_unipolar_ao;
	int ai_fifo_depth;
};

static struct ni_boardtype_struct ni_boardtypes[]={
	{44,"at-mio-16e-1"  ,16,12,0,0,1,1,2,1,8192 },
	{25,"at-mio-16e-2"  ,16,12,0,0,1,1,2,1,2048 },
	{36,"at-mio-16e-10" ,16,12,0,0,1,2,2,1,512 },
	{37,"at-mio-16de-10",16,12,0,0,1,2,2,1,512 },
	{38,"at-mio-64e-3"  ,64,12,0,0,1,1,2,1,2048 },
	{39,"at-mio-16xe-50",16,16,1,1,2,3,2,0,512 },
	{00,"at-mio-16xe-10",16,16,1,0,2,3,2,1,512 },	/* unknown */
	{00,"at-ai-16xe-10",16,16,1,0,2,3,0,0,512 }	/* unknown */
};
#define N_BOARDTYPES 8

static int ni_gainlkup[]={ 1, 2, 4, 7, 9, 10, 12, 15 };
static int ni_irqpin[]={-1,-1,-1,0,1,2,-1,3,-1,-1,4,5,6,-1,-1,7};

/* How we access windowed registers on the AT-MIO */

#define _win_out(a,b,c) (outw((b),(c)+Window_Address),outw((a),(c)+Window_Data))
#define win_out(a,b) _win_out((a),(b),dev->iobase)
#define _win_in(b,c) (outw((b),(c)+Window_Address),inw((c)+Window_Data))
#define win_in(b) _win_in((b),dev->iobase)

/* Additional windowed registers unique to E series */

#define Configuration_Memory_Clear	82
#define ADC_FIFO_Clear			83


/* i/o port offsets */

/* 8 bit registers */
#define Status				0x01
#define Serial_Command			0x0d
#define Misc_Command			0x0f
#define Port_A				0x19
#define Port_B				0x1b
#define Port_C				0x1d
#define Configuration			0x1f
#define Strobes				0x01
#define Channel_A_Mode			0x03
#define Channel_B_Mode			0x05
#define Channel_C_Mode			0x07
#define AI_AO_Select			0x09
#define G0_G1_Select			0x0b

/* 16 bit registers */
#define Configuration_Memory_Low	0x10
#define Configuration_Memory_High	0x12
#define ADC_FIFO_Data_Register		0x1c
#define AO_Configuration		0x16
#define DAC_FIFO_Data			0x1e
#define DAC0_Direct_Data		0x18
#define DAC1_Direct_Data		0x1a


typedef struct{
	int board;
	int dio;
	int ao0p,ao1p;
	int aip[64];
	int lastchan;
	int last_do;
}ni_private;
#define devpriv ((ni_private *)dev->private)
#define boardtype ni_boardtypes[devpriv->board]

comedi_devinit ni_E_init;
static int ni_dio(comedi_device *dev,comedi_trig *ad);
static int ni_ai(comedi_device *dev,comedi_trig *it);
static int ni_rem(comedi_device *dev);
/*static int pnp_init(int iobase);*/
static int ni_read_eeprom(comedi_device *dev,int addr);
static int ni_getboardtype(comedi_device *dev);
static void ni_calibrate(comedi_device *dev);
static int ni_memcpy_from(comedi_device *dev,comedi_memcpy *it);

static void ni_E_interrupt(int irq,void *d,struct pt_regs * regs)
{
	comedi_sample s;
	comedi_device *dev=d;
	int wake=0 ;
	int status;

/*
    If you want to use windowed registers in an interrupt, it is
    important that you restore the window address register.  If
    you change certain modes, e.g., AI_Configuration_Start/End,
    you need to set up software flags for non-interrupt routines.
*/
	
	status=inw(AI_Status_1+dev->iobase);
	if(status&(AI_Overrun_St|AI_Overflow_St))
		printk("ni_E: overrun/overflow status=0x%04x\n",status);
	
	if(status&AI_FIFO_Half_Full_St){
	/*
	   if we got a fifo_half_full interrupt, we can transfer fifo/2
	   samples witout checking the empty flag.  It doesn't matter if
	   we transfer the rest of the samples, the performance trade-off
	   is minimal (checking empty flag for a few samples vs. having
	   1% more interrupts.)  At really high speeds, it's better to
	   ignore them.
	*/
		int i;
	
		for(i=0;i<boardtype.ai_fifo_depth/2;i++){
			s.data=inw(dev->iobase+ADC_FIFO_Data_Register);
			if(!(devpriv->aip[devpriv->lastchan]&0x100))
				s.data^=0x800;
			s.data&=0xfff; /* incompatible with 16 bit cards */
			s.chan=devpriv->lastchan;
			report_sample(dev,&s);
		}
		return;
	}
	if(status&AI_SC_TC_St){
	/* 
	    end of sampling.  Read the rest of the FIFO.
	*/
		while(!(inw(AI_Status_1+dev->iobase)&AI_FIFO_Empty_St)){
			s.data=inw(dev->iobase+ADC_FIFO_Data_Register);
			if(!(devpriv->aip[devpriv->lastchan]&0x100))
				s.data^=0x800;
			s.data&=0xfff; /* incompatible with 16 bit cards */
			s.chan=devpriv->lastchan;
			report_sample(dev,&s);
			wake=1;
		}
		dev->busy[0]=0;
	}
	if(wake){
		wake_up_interruptible(&dev->rdwait);
	}
}

#if 0
static int ni_stop_job(comedi_device *dev,int job)
{
/* assume it means ai job*/
	win_out(AI_Configuration_Start,Joint_Reset_Register);
	win_out(AI_Configuration_End,Joint_Reset_Register);

	return 0;
}
#endif

static int ni_iai(comedi_device *dev,comedi_trig *it)
{
	int mode,chan;
	int i;
	
#ifdef TRUX
	/* this is a hack, but then, so is busy[], so I'll allow it for a while   -ds */
	/* special case : kill job */
	if(it->trigsrc==COMEDI_TRIGRESET)
	  {
	    ni_stop_job(dev,dev->busy[0]) ;
	    dev->busy[0]=0 ;
	    return(0) ;
	  }

#endif /* TRUX */

	if(dev->busy[0])return -EBUSY;
	
	win_out(1,ADC_FIFO_Clear);
	win_out(1,Configuration_Memory_Clear);

	/* interrupt on errors */
	win_out(0x0020,Interrupt_A_Enable_Register) ;
	
	mode=getparamval(dev->chinfo[it->chan].paramlist,COMEDI_AREF);
	chan=ni_modebits1[mode]|(it->chan&ni_modebits2[mode]);

	outw(chan,dev->iobase+Configuration_Memory_High);
	outw(0x8000|devpriv->aip[it->chan],
		dev->iobase+Configuration_Memory_Low);
	win_out(1,AI_Command_1_Register);
	for(i=0;i<40;i++){
		if(!(inw(AI_Status_1+dev->iobase)&
		  AI_FIFO_Empty_St)){
			win_out(1,ADC_FIFO_Clear);
			break;
		}
		udelay(25);
	}
	if(i==40){
		printk("ni_E: timeout 1\n");
		return -ETIME;
	}

	win_out(1,AI_Command_1_Register);
	
	for(i=0;i<40;i++){
		if(!(inw(AI_Status_1+dev->iobase)&
		  AI_FIFO_Empty_St)){
			it->data=inw(dev->iobase+ADC_FIFO_Data_Register);
			if(boardtype.adbits==12){
				if(!(devpriv->aip[it->chan]&0x100))
					it->data^=0x800;
				it->data&=0xfff;
			}else{
				if(!(devpriv->aip[it->chan]&0x100))
					it->data^=0x8000;
				it->data&=0xffff;
			}
			return 0;
		}
		udelay(25);
	}
	printk("ni_E: timeout 2\n");
	return -ETIME;
}

static int ni_ai(comedi_device *dev,comedi_trig *it)
{
	int mode,chan;
	
	if(dev->busy[0])return -EBUSY;
	switch(it->trigsrc){
	case COMEDI_TRIGCLK:
		/* use FIFO interrupts to clean FIFO */
		devpriv->lastchan=it->chan;
		
		win_out(1,ADC_FIFO_Clear);
		win_out(1,Configuration_Memory_Clear);
	
		mode=getparamval(dev->chinfo[it->chan].paramlist,COMEDI_AREF);
		chan=ni_modebits1[mode]|(it->chan&ni_modebits2[mode]);
	
		outw(chan,dev->iobase+Configuration_Memory_High);
		outw(0x8000|devpriv->aip[it->chan],
			dev->iobase+Configuration_Memory_Low);
		
		/* start configuration */
		win_out(AI_Configuration_Start,Joint_Reset_Register);
	
		/* stage number of samples */
		win_out((it->n-1)>>16,AI_SC_Load_A_Registers);
		win_out((it->n-1)&0xffff,AI_SC_Load_A_Registers+1);
		
		/* load SC */
		win_out(0x20,AI_Command_1_Register);
		
		/*
		    AI_SI_Special_Trigger_Delay=0
		    AI_Pre_Trigger=0
		  AI_START_STOP_Select_Register:
		    AI_START_Polarity=0 (?)	rising edge
		    AI_START_Edge=1		edge triggered
		    AI_START_Sync=1 (?)		
		    AI_START_Select=0		SI_TC
		    AI_STOP_Polarity=0		rising edge
		    AI_STOP_Edge=0		level
		    AI_STOP_Sync=1		
		    AI_STOP_Select=19		external pin (configuration mem)
		win_out(AI_START_Edge|AI_START_Sync|AI_STOP_Select(19)|
			AI_STOP_Sync,AI_START_STOP_Register);
		*/
#if 0
/* deprecated code -- doesn't work */
		/* not pretriggered acquisition */
		/* win_out(0x0000,AI_Mode_2_Register); */
	
		/* AI_SI_Special_Trigger_Delay=0 */
		/* win_out(0,AI_Mode_3_Register); */
		
		/* AI_START_Edge, AI_START_Sync =1 */
		/* win_out(0x60,AI_START_STOP_Select_Register); */
#define AI_START_Polarity		_bit15
#define AI_STOP_Polarity		_bit14
#define AI_STOP_Sync			_bit13
#define AI_STOP_Edge			_bit12
#define AI_STOP_Select(a)		((a)<<7)
#define AI_START_Sync			_bit6
#define AI_START_Edge			_bit5
#define AI_START_Select(a)		(a)
win_out(AI_START_Edge|AI_START_Sync|AI_STOP_Select(19)|
	AI_STOP_Sync,AI_START_STOP_Select_Register);
		
		/* AI_SI_Source_Select, AI_SI_Source_Polarity */
		/* win_out(0x0000,AI_Mode_1_Register); */
	
		win_out((it->trigvar>>16),AI_SI_Load_B_Registers);
		win_out((it->trigvar&0xffff),AI_SI_Load_B_Registers+1);
		
		win_out((it->trigvar>>16),AI_SI_Load_A_Registers);
		win_out((it->trigvar&0xffff),AI_SI_Load_A_Registers+1);
		
		/* initial SI load register = B, always use B */
		win_out(0x80,AI_Mode_2_Register);
		
		/* load SI */
		win_out(0x200,AI_Command_1_Register);
		
		/* alternate 1st period on every SC_TC */
		win_out(6<<4,AI_Mode_2_Register);

		/*win_out((0x60)|(31<<7)|(1<<14),AI_START_STOP_Select_Register);*/
		
		/* alternate 1st period on every SC_TC */
		/* AI_SC_Gate_Enable=0 */
		/* AI_Start_Stop_Gate_Enable=0 */
		win_out((6<<4)|(0<<15)|(0<<14),AI_Mode_2_Register);
		
		/*win_out(0<<11,AI_Mode_3_Register);*/
		win_out(AI_Configuration_End,Joint_Reset_Register);
		
		win_out(1,ADC_FIFO_Clear);
#else
#define AI_SC_Gate_Enable		_bit15
#define AI_Start_Stop_Gate_Enable	_bit14
#define AI_Pre_Trigger			_bit13
#define AI_External_MUX_Present		_bit12
#define AI_SI2_Ininial_Load_Source	_bit9
#define AI_SI2_Reload_Mode		_bit8
#define AI_SI_Initial_Load_Source	_bit7
#define AI_SI_Reload_Mode(a)		((a)<<4)
#define AI_SI_Write_Switch		_bit3
#define AI_SC_Initial_Load_Source	_bit2
#define AI_SC_Reload_Mode		_bit1
#define AI_SC_Write_Switch		_bit0
#if 0
		/* stage 1 into SI A */
		win_out(0,AI_SI_Load_A_Registers);
		win_out(1,AI_SI_Load_A_Registers+1); */
#endif
		win_out((it->trigvar>>16),AI_SI_Load_A_Registers);
		win_out((it->trigvar&0xffff),AI_SI_Load_A_Registers+1);
		/* AI_SI_Initial_Load_Source=A */
		win_out(0,AI_Mode_2_Register);
		/* load SI */
		win_out(0x200,AI_Command_1_Register);

		/* stage freq. counter into SI B */
		win_out((it->trigvar>>16),AI_SI_Load_B_Registers);
		win_out((it->trigvar&0xffff),AI_SI_Load_B_Registers+1);
	
		win_out(1,AI_SI2_Load_A_Register); /* 0,0 does not work. */
		win_out(1,AI_SI2_Load_B_Register);

		/* AI_SI2_Reload_Mode = alternate */
		/* AI_SI2_Initial_Load_Source = A */
		win_out(0x0100,AI_Mode_2_Register);
	
		/* AI_SI2_Load */
		win_out(0x0800,AI_Command_1_Register);
	
		/* AI_SI_Initial_Load_Source=0
		   AI_SI_Reload_Mode(0)
		   AI_SI2_Reload_Mode = alternate, AI_SI2_Initial_Load_Source = B */
		win_out(0x0300,AI_Mode_2_Register);

		/* end configuration */
		win_out(AI_Configuration_End,Joint_Reset_Register);

#if 0
/* this is now set in init() */
		/* set FIFO to interrupt on half full */
		win_out((1<<6),AI_Mode_3_Register);
#endif

#endif
		if(dev->irq){
			/* interrupt on FIFO, errors, SC_TC */
			win_out(0x00a1,Interrupt_A_Enable_Register) ;
		}else{
			/* interrupt on nothing */
			win_out(0x0000,Interrupt_A_Enable_Register) ;
		}
		
		/* AI_SI2_Arm, AI_SI_Arm, AI_DIV_Arm, AI_SC_Arm */
		win_out(0x1540,AI_Command_1_Register);

		/* AI_START1_Pulse */
		win_out(1,AI_Command_2_Register);

		dev->busy[0]=it->job;
		if(!dev->irq){
			start_polling(dev);
		}
		return 0;
	default:
		return -EINVAL;
	}
}

static int ni_aisp(comedi_device *dev,comedi_param *it)
{
	switch(it->pnum){
	case COMEDI_RANGE:
	{
		int pval=it->pval;
		
		if(pval<0 || pval>=16)return -EINVAL;
		if(boardtype.limitedgain){
			if(pval>=8)return -EINVAL;
			pval=ni_gainlkup[it->pval];
		}
		changeparam(&dev->chinfo[it->chan].paramlist,COMEDI_RANGE,
			0,it->pval);
		devpriv->aip[it->chan]&=~0x107;
		devpriv->aip[it->chan]|=((pval&0x8)<<5);
		devpriv->aip[it->chan]|=(pval&0x7);
		return 0;
	}
	case COMEDI_AREF:
		if(it->pval<0 || it->pval>=4)return -EINVAL;
		changeparam(&dev->chinfo[it->chan].paramlist,COMEDI_AREF,
			0,it->pval);
		return 0;
	case COMEDI_DITHER:
		if(boardtype.alwaysdither)return -EINVAL;
		if(it->pval<0 || it->pval>=2)return -EINVAL;
		changeparam(&dev->chinfo[it->chan].paramlist,COMEDI_DITHER,
			0,it->pval);
		devpriv->aip[it->chan]&=~0x200;
		devpriv->aip[it->chan]|=(it->pval<<9);
		return 0;
	default:
		return -EINVAL;
	}
}

static int ni_ao(comedi_device *dev,comedi_trig *it)
{
	unsigned int i=it->data;
	
	if(i>=0x1000)return -EINVAL;
	i^=0x800;
	if(it->chan&1){
		if(!(devpriv->ao0p&1))i^=0x800;
		outw(i,dev->iobase+DAC0_Direct_Data);
	}else{
		if(!(devpriv->ao1p&1))i^=0x800;
		outw(i,dev->iobase+DAC1_Direct_Data);
	}
	return 0;
}

static int ni_aosp(comedi_device *dev,comedi_param *it)
{
	switch(it->pnum){
	case COMEDI_RANGE:
		if(it->pval<0 || it->pval>=2)return -EINVAL;
		changeparam(&dev->chinfo[it->chan].paramlist,COMEDI_RANGE,
			0,it->pval);
		if(it->chan==16){
			devpriv->ao0p&=~1;
			devpriv->ao0p|=(it->pval?1:0);
			outw(devpriv->ao0p,dev->iobase+AO_Configuration);
		}else{
			devpriv->ao1p&=~1;
			devpriv->ao1p|=(it->pval?1:0);
			outw(devpriv->ao1p,dev->iobase+AO_Configuration);
		}
		return 0;
	case COMEDI_DEGLITCH:
		if(it->pval<0 || it->pval>=2)return -EINVAL;
		changeparam(&dev->chinfo[it->chan].paramlist,COMEDI_DEGLITCH,
			0,it->pval);
		if(it->chan==16){
			devpriv->ao0p&=~2;
			devpriv->ao0p|=(it->pval?2:0);
			outw(devpriv->ao0p,dev->iobase+AO_Configuration);
		}else{
			devpriv->ao1p&=~2;
			devpriv->ao1p|=(it->pval?2:0);
			outw(devpriv->ao1p,dev->iobase+AO_Configuration);
		}
		return 0;
	default:
		return -EINVAL;
	}
}

static int ni_dio(comedi_device *dev,comedi_trig *it)
{
	int data,mask;

	switch(it->io){
	case COMEDI_IO_INPUT:
		it->data= win_in(DIO_Input_Register)&0xff;
	
		return 0;
	case COMEDI_IO_OUTPUT:
		if(it->n)mask=it->n; else mask=0xff;
		data=(devpriv->last_do&~mask)|(it->data&mask);
		win_out(data,DIO_Output_Register);
		devpriv->last_do=data;
		return 0;
	default:
		return -EINVAL;
	}
}

static int ni_dsp(comedi_device *dev,comedi_param *it)
{
	switch(it->pnum){
	case COMEDI_IOBITS:
		if(it->pval&~0xff)return -EINVAL;
		devpriv->dio=it->pval;
		win_out(devpriv->dio,DIO_Control_Register);
		changeparam(&dev->chinfo[it->chan].paramlist,COMEDI_IOBITS,0,
			it->pval);
		return 0;
	default:
		return -EINVAL;
	}
}

/*
   XXX this assumes that if there isn't anything in the fifo,
   then we can turn off polling.  this is *not* a good idea,
   but it will do for now.
   Note: there is always the possibility that the poll
   routine will steal a sample from a soft triggered AI.
*/

static void ni_E_poll(comedi_device *dev)
{
	comedi_sample s;
	int wake=0;
	int i;
	int status;
#ifdef DEBUG
	int n=0;
#endif
	
	status=inw(AI_Status_1+dev->iobase);
#ifdef DEBUG
	printk("ni_E: polling... jiffies=%d status=0x%04x ",jiffies,status);
#endif

	if(status&AI_FIFO_Half_Full_St){
#ifdef DEBUG
		printk("fifo half full ");
		n+=boardtype.ai_fifo_depth/2;
#endif
		for(i=0;i<boardtype.ai_fifo_depth/2;i++){
			s.data=inw(dev->iobase+ADC_FIFO_Data_Register);
			if(!(devpriv->aip[devpriv->lastchan]&0x100))
				s.data^=0x800;
			s.data&=0xfff; /* incompatible with 16 bit cards */
			s.chan=devpriv->lastchan;
			report_sample(dev,&s);
		}
		wake=1;
	}
	while(!(inw(AI_Status_1+dev->iobase)&AI_FIFO_Empty_St)){
		s.data=inw(dev->iobase+ADC_FIFO_Data_Register);
		if(!(devpriv->aip[devpriv->lastchan]&0x100))
			s.data^=0x800;
		s.data&=0xfff; /* incompatible with 16 bit cards */
		s.chan=devpriv->lastchan;
		report_sample(dev,&s);
		wake=1;
#ifdef DEBUG
		n++;
#endif
	}
#ifdef DEBUG
	printk("samples=%d\n",n);
#endif
	if(wake)wake_up_interruptible(&dev->rdwait);
	else{
		stop_polling(dev);
		dev->busy[0]=0;
	}
}

/*
	options:
	0	I/O base
	1	irq to use (0 for none)
*/

int ni_E_init(comedi_device *dev,comedi_devconfig *it)
{
	int i,board,irq ;

	if(strcmp("ni_E",it->name))
		return 0;

	if(it->options[0])dev->iobase=it->options[0];
	else dev->iobase=0x200;
	
	printk("comedi%d: ni_E: 0x%04x",dev->minor,dev->iobase);
	if(check_region(dev->iobase,NI_SIZE)<0){
		printk(" I/O port conflict\n");
		return -EIO;
	}
	dev->iosize=NI_SIZE;
	
	dev->irq=0;
	irq=it->options[1];
	if(irq!=0){
		if(irq==(irq&0xf) && ni_irqpin[irq] != -1){
			printk("( irq = %d )\n",irq);
			request_irq(irq,ni_E_interrupt,SA_INTERRUPT,"ni_E",dev);
			dev->irq=irq;
		}else return -EINVAL;
	}

	board=ni_getboardtype(dev);
	if(board<0)return -EIO;
	
	printk(" %s",ni_boardtypes[board].name);
	dev->board_name=ni_boardtypes[board].name;
	dev->driver_name="ni_E";

	dev->chan=ni_boardtypes[board].n_adchan+3;

	devpriv=kmalloc(sizeof(ni_private)+dev->chan*sizeof(comedi_chinfo),
		GFP_KERNEL);
	if(!devpriv){
		return -ENOMEM;
	}
	memset(devpriv,0,sizeof(ni_private)+dev->chan*sizeof(comedi_chinfo));
	devpriv->board=board;
	request_region(dev->iobase,NI_SIZE,"ni_E");
	
	dev->chinfo=(void *)(devpriv+1);
	for(i=0;i<boardtype.n_adchan;i++){
		dev->chinfo[i].itrig=ni_iai;
		dev->chinfo[i].trig=ni_ai;
		dev->chinfo[i].sp=ni_aisp;
		addparam(&dev->chinfo[i].paramlist,
			COMEDI_BITS,COMEDI_RDONLY,boardtype.adbits);
		addparam(&dev->chinfo[i].paramlist,
			COMEDI_MAXDATA,COMEDI_RDONLY,(1<<boardtype.adbits)-1);
		addparam(&dev->chinfo[i].paramlist,
			COMEDI_AD,COMEDI_RDONLY,COMEDI_AD_ANALOG);
		addparam(&dev->chinfo[i].paramlist,
			COMEDI_IO,COMEDI_RDONLY,COMEDI_IO_INPUT);
		addparam(&dev->chinfo[i].paramlist,COMEDI_RANGE,0,0);
		addparam(&dev->chinfo[i].paramlist,COMEDI_RANGE_TYPE,
			COMEDI_RDONLY,boardtype.limitedgain?
			RANGE_ni_E_ai_limited:RANGE_ni_E_ai);
		addparam(&dev->chinfo[i].paramlist,COMEDI_AREF,COMEDI_RDWR,
			COMEDI_AREF_DIFF);
		if(boardtype.alwaysdither){
			addparam(&dev->chinfo[i].paramlist,COMEDI_DITHER,
				COMEDI_RDONLY,1);
			devpriv->aip[i]=0x0100;
		}else{
			addparam(&dev->chinfo[i].paramlist,COMEDI_DITHER,
				COMEDI_RDWR,0);
			devpriv->aip[i]=0x0000;
		}
	}
	dev->chinfo[i].itrig=ni_ao;
	dev->chinfo[i].sp=ni_aosp;
	addparam(&dev->chinfo[i].paramlist,
		COMEDI_BITS,COMEDI_RDONLY,12);
	addparam(&dev->chinfo[i].paramlist,
		COMEDI_MAXDATA,COMEDI_RDONLY,(1<<12)-1);
	addparam(&dev->chinfo[i].paramlist,
		COMEDI_AD,COMEDI_RDONLY,COMEDI_AD_ANALOG);
	addparam(&dev->chinfo[i].paramlist,
		COMEDI_IO,COMEDI_RDONLY,COMEDI_IO_OUTPUT);
	addparam(&dev->chinfo[i].paramlist,COMEDI_RANGE_TYPE,
		COMEDI_RDONLY,RANGE_bipolar10);
	addparam(&dev->chinfo[i].paramlist,COMEDI_RANGE,0,0);
	addparam(&dev->chinfo[i].paramlist,COMEDI_DEGLITCH,0,0);
	i++;
	
	dev->chinfo[i].itrig=ni_ao;
	dev->chinfo[i].sp=ni_aosp;
	addparam(&dev->chinfo[i].paramlist,
		COMEDI_BITS,COMEDI_RDONLY,12);
	addparam(&dev->chinfo[i].paramlist,
		COMEDI_MAXDATA,COMEDI_RDONLY,0xfff);
	addparam(&dev->chinfo[i].paramlist,
		COMEDI_AD,COMEDI_RDONLY,COMEDI_AD_ANALOG);
	addparam(&dev->chinfo[i].paramlist,
		COMEDI_IO,COMEDI_RDONLY,COMEDI_IO_OUTPUT);
	addparam(&dev->chinfo[i].paramlist,COMEDI_RANGE_TYPE,
		COMEDI_RDONLY,RANGE_bipolar10);
	addparam(&dev->chinfo[i].paramlist,COMEDI_RANGE,0,0);
	addparam(&dev->chinfo[i].paramlist,COMEDI_DEGLITCH,0,0);
	i++;

	dev->chinfo[i].itrig=ni_dio;
	dev->chinfo[i].sp=ni_dsp;
	addparam(&dev->chinfo[i].paramlist,
		COMEDI_BITS,COMEDI_RDONLY,8);
	addparam(&dev->chinfo[i].paramlist,
		COMEDI_AD,COMEDI_RDONLY,COMEDI_AD_DIGITAL);
	addparam(&dev->chinfo[i].paramlist,
		COMEDI_IO,COMEDI_RDONLY,COMEDI_IO_BIDIR);
	addparam(&dev->chinfo[i].paramlist,
		COMEDI_IOBITGROUP,COMEDI_RDONLY,1);
	addparam(&dev->chinfo[i].paramlist,
		COMEDI_IOBITS,COMEDI_RDWR,0xff&COMEDI_IOBITS_INPUT);
	devpriv->dio=0;
	win_out(devpriv->dio,DIO_Control_Register);

#ifndef USE_OLD_INITIALIZATION
	win_out(AI_Configuration_Start,Joint_Reset_Register);

	win_out(0x3f80,Interrupt_A_Ack_Register); /* clear interrupts */
	win_out(0x0000,AI_Command_1_Register); /* reset pulses */
	win_out(0x1ba0,Clock_and_FOUT_Register);
	win_out(0x000d,AI_Mode_1_Register);
	win_out(0x0000,AI_Mode_2_Register);
	win_out((1<<6)|0x0000,AI_Mode_3_Register);
	win_out(0xa4a0,AI_Personal_Register); /* ? */
	win_out(0x032e,AI_Output_Control_Register);
	win_out(0x0060,AI_Trigger_Select_Register); /* trigger source */
	win_out(0x29e0,AI_START_STOP_Select_Register);

	win_out(AI_Configuration_End,Joint_Reset_Register);
	
	/* the following registers should not be changed:
		Clock_and_FOUT_Register
		AI_Mode_1_Register
		AI_Mode_3_Register
		AI_Personal_Register
		AI_Output_Control_Register
		AI_Trigger_Select_Register
		AI_START_STOP_Select_Register
	*/

#else
	win_out(1,ADC_FIFO_Clear);
	win_out(AI_Reset,Joint_Reset_Register);  /* remove? */
	win_out(1,Configuration_Memory_Clear);  /* remove? */
	win_out(1,ADC_FIFO_Clear);  /* remove? */
	win_out(0x1b00,Clock_and_FOUT_Register); /* set timebase */
	win_out(AI_Configuration_Start,Joint_Reset_Register);
	win_out(0x3f80,Interrupt_A_Ack_Register); /* clear interrupts */
	win_out(0x0000,AI_Command_1_Register); /* don't do anything */
	win_out(0x000c,AI_Mode_1_Register); /* listen to start/stop cmds */
	win_out(AI_Configuration_End,Joint_Reset_Register);  /* remove? */
	win_out(AI_Configuration_Start,Joint_Reset_Register);  /* remove? */
	win_out(0x1b80,Clock_and_FOUT_Register); /* move? */
	win_out(0xa4a0,AI_Personal_Register); /* ? */
	win_out(0x032e,AI_Output_Control_Register); /* output signals */
	win_out(0x032e,AI_Output_Control_Register); /* remove? */
	win_out(AI_Configuration_End,Joint_Reset_Register);  /* remove? */
	win_out(AI_Configuration_Start,Joint_Reset_Register);  /* remove? */
	win_out(0,AI_Mode_2_Register); /* ? */
	win_out(0x032e,AI_Output_Control_Register); /* remove? */
	win_out(AI_Configuration_End,Joint_Reset_Register);  /* remove? */
	
	/* generate FIFO interrupts on half full */
	win_out((1<<6),AI_Mode_3_Register); /* ? */

	win_out(AI_Configuration_Start,Joint_Reset_Register);  /* remove? */
	win_out(0x000d,AI_Mode_1_Register); /* move? */
	win_out(0x0060,AI_Trigger_Select_Register); /* trigger source */
	win_out(AI_Configuration_End,Joint_Reset_Register);  /* remove? */
	win_out(AI_Configuration_Start,Joint_Reset_Register);  /* remove? */
	win_out(0x0060,AI_START_STOP_Select_Register); /* remove? */
	win_out(AI_Configuration_End,Joint_Reset_Register);  /* remove? */
	win_out(AI_Configuration_Start,Joint_Reset_Register);  /* remove? */
		/* start on AI_START1_Pulse */
		/* stop on AI_STOP_IN pin */
	win_out(0x29e0,AI_START_STOP_Select_Register);
	win_out(AI_Configuration_End,Joint_Reset_Register);
	win_out(1,ADC_FIFO_Clear);  /* remove? */
#endif
	
	devpriv->ao0p=0x0000;
	outw(devpriv->ao0p,dev->iobase+AO_Configuration);
	devpriv->ao1p=0x0100;
	outw(devpriv->ao1p,dev->iobase+AO_Configuration);
	win_out(AO_Configuration_Start,Joint_Reset_Register);
	win_out(0x2000,AO_Command_1_Register);
	win_out(0,Interrupt_B_Enable_Register);
	win_out(0x0010,AO_Personal_Register);
	outw(0x3f98,dev->iobase+Interrupt_B_Ack);
	win_out(0x1430,AO_Personal_Register);
	win_out(0,AO_Output_Control_Register);
	win_out(0,AO_Start_Select_Register);
	win_out(0,AO_Command_1_Register);
	if(dev->irq)
		win_out((0<<0) |  /* polarity : active high */
			(0<<1) |  /* no interrupt on 3 pins */
			(1<<11) |  /* enable int A */
			(ni_irqpin[dev->irq]<<8),  /* interrupt output pin */
			Interrupt_Control_Register
			);

	ni_calibrate(dev);
	
	comedi_initbuf(&dev->rdbuf);

	dev->poll=ni_E_poll;
	dev->memcpy_from=ni_memcpy_from;
	dev->rem=ni_rem;

	printk("\n");

	return 1;
}


static int ni_rem(comedi_device *dev)
{
	int i;

	printk("comedi%d: ni_E: remove\n",dev->minor);
	
	for(i=0;i<dev->chan;i++)
		delparamlist(dev->chinfo[i].paramlist);
	kfree(devpriv);
	release_region(dev->iobase,dev->iosize);
	if(dev->irq)free_irq(dev->irq,dev);

	comedi_freebuf(&dev->rdbuf);

	return 0;
}

static int ni_read_eeprom(comedi_device *dev,int addr)
{
	int bit;
	int bitstring;

	bitstring=0x0300|((addr&0x100)<<3)|(addr&0xff);
	outb(0x04,dev->iobase+Serial_Command);
	for(bit=0x8000;bit;bit>>=1){
		outb(0x04|((bit&bitstring)?0x02:0),dev->iobase+Serial_Command);
		outb(0x05|((bit&bitstring)?0x02:0),dev->iobase+Serial_Command);
	}
	bitstring=0;
	for(bit=0x80;bit;bit>>=1){
		outb(0x04,dev->iobase+Serial_Command);
		outb(0x05,dev->iobase+Serial_Command);
		bitstring|=((inb(dev->iobase+Status)&0x01)?bit:0);
	}
	outb(0x00,dev->iobase+Serial_Command);
	
	return bitstring;
}

static void ni_write_caldac(comedi_device *dev,int addr,int val)
{
	int loadbit,bits,bit,bitstring;
	
	switch(boardtype.caldactype){
	case 1:
		/*
		   MB 88341
		   Note that address bits are reversed.  Thanks to
		   Ingo Keen for noticing this.
		*/
		bits=12;
		bitstring=((addr&0x1)<<11) |
			  ((addr&0x2)<<9)  |
			  ((addr&0x4)<<7)  |
			  ((addr&0x8)<<5)  |
			  (val&0xff);
		loadbit=0x08;
		break;
	case 2:
		if(addr==8){
			/*
			   DAC 8043
			*/
			bits=12;
			bitstring=val&0xfff;
			loadbit=0x10;
		}else{
			/*
			   DAC 8800
			*/
			bits=11;
			bitstring=((addr&0x7)<<8)|(val&0xff);
			loadbit=0x08;
		}
		break;
	default:
		return;
	}

	for(bit=1<<(bits-1);bit;bit>>=1){
		outb(((bit&bitstring)?0x02:0),dev->iobase+Serial_Command);
		outb(1|((bit&bitstring)?0x02:0),dev->iobase+Serial_Command);
	}
	outb(loadbit,dev->iobase+Serial_Command);
	outb(0,dev->iobase+Serial_Command);
}

static int ni_getboardtype(comedi_device *dev)
{
	int boardcode=ni_read_eeprom(dev,511);
	int i;
	
	for(i=0;i<N_BOARDTYPES;i++){
		if(ni_boardtypes[i].boardcode==boardcode){
			return i;
		}
	}
	printk(" unknown board code %d -- contact author\n",boardcode);
	return -1;
}

static void ni_calibrate(comedi_device *dev)
{
	switch(boardtype.eepromlayout){
	case 1:
		ni_write_caldac(dev,4,ni_read_eeprom(dev,424));
		ni_write_caldac(dev,1,ni_read_eeprom(dev,423));
		ni_write_caldac(dev,3,ni_read_eeprom(dev,422));
		ni_write_caldac(dev,2,ni_read_eeprom(dev,421));
		break;
	case 2:
		ni_write_caldac(dev,4,ni_read_eeprom(dev,422));
		ni_write_caldac(dev,11,ni_read_eeprom(dev,421));
		ni_write_caldac(dev,1,ni_read_eeprom(dev,420));
		ni_write_caldac(dev,3,ni_read_eeprom(dev,419));
		ni_write_caldac(dev,2,ni_read_eeprom(dev,418));
		break;
	default:
	}
}

static int ni_memcpy_from(comedi_device *dev,comedi_memcpy *it)
{
	char *l=it->local;
	int i;
	
	if(it->remote<0 || it->nbytes<0 ||
	  it->remote+it->nbytes>512)return -EINVAL;
	for(i=0;i<it->nbytes;i++){
		l[i]=ni_read_eeprom(dev,it->remote+i);
	}
	return 0;
}


/*********************************************/

typedef struct{
	int device_id;
	char *name;
}ni_pci_board;

#define PCI_VENDOR_ID_NATINST		0x1093

ni_pci_board ni_pci_boards[]={
	{ 0x0162, "pci-mio-16xe-50" },
	{ 0x1170, "pci-mio-16xe-10" },
	{ 0x1180, "pci-mio-16e-1" },
	{ 0x1190, "pci-mio-16e-4" },
};
int n_ni_pci_boards = sizeof(ni_pci_boards)/sizeof(ni_pci_board);


void config_pci_device(void)
{
	struct pci_dev	*dev=NULL;
	int i;
	
	if(!pci_present())
		return 0;
	
	for(i=0;i<
	while((dev=pci_find_device(vendor_id,device_id,dev)){
		
	}
	
	find_card();
	setup_mite();
}




void find_ni_devices(void)
{

}

setup_mite()
{

}

unset_mite()
{

}

