/*
    module/pcl711.c
    hardware driver for PC-LabCard PCL-711

    COMEDI - Linux Control and Measurement Device Interface
    Copyright (C) 1998 David A. Schleef <ds@stm.lbl.gov>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

/*
	Dave Andruczyk <xxx@tech.buffalostate.edu> also wrote a
	driver for this card.  I used a few ideas from his driver
	here.  His driver also has more comments, if you are
	interested in understanding how this driver works.
*/

#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/mm.h>
#include <linux/malloc.h>
#include <linux/errno.h>
#include <linux/ioport.h>
#include <linux/delay.h>
#include <linux/interrupt.h>
#include <linux/timex.h>
#include <linux/timer.h>
#include <asm/io.h>
#include <comedi_module.h>



#define PCL711_SIZE 16

#define PCL711_CTR0 0
#define PCL711_CTR1 1
#define PCL711_CTR2 2
#define PCL711_CTRCTL 3
#define PCL711_AD_LO 4
#define PCL711_DA_LO 4
#define PCL711_AD_HI 5
#define PCL711_DA_HI 5
#define PCL711_DI_LO 6
#define PCL711_DI_HI 7
#define PCL711_CLRINTR 8
#define PCL711_GAIN 9
#define PCL711_MUX 10
#define PCL711_MODE 11
#define PCL711_SOFTTRIG 12
#define PCL711_DO_LO 13
#define PCL711_DO_HI 14

/*
 * flags
 */

#define PCL711_TIMEOUT 100
#define PCL711_DRDY 0x10

comedi_devinit pcl711_init;
static int pcl711_rem(comedi_device *dev);

typedef struct{
	int is_pcl711b;
	int adchan;
	int ntrig;
	int aip[8];
	int last_do;
}pcl711_private;
#define devpriv ((pcl711_private *)dev->private)

static void pcl711_interrupt(int irq,void *d,struct pt_regs *regs)
{
	int lo,hi;
	comedi_device *dev=d;
	comedi_sample pt;
	
	hi=inb(dev->iobase+PCL711_AD_HI);
	lo=inb(dev->iobase+PCL711_AD_LO);
	outb(0,dev->iobase+PCL711_CLRINTR);
	
	pt.data=(hi<<8)|lo;
	pt.chan=devpriv->adchan;
	report_sample(dev,&pt);
	
	if(!(--devpriv->ntrig)){
		outb(0,dev->iobase+PCL711_MODE);
		dev->busy[0]=0;
		
		wake_up_interruptible(&dev->rdwait);
	}
}
	

static int pcl711_iai(comedi_device *dev,comedi_trig *it)
{
	int hi,lo,i;

	if(dev->busy[0])return -EBUSY;
	outb(devpriv->aip[it->chan],dev->iobase+PCL711_GAIN);
	outb(it->chan,dev->iobase+PCL711_MUX);
	outb(0,dev->iobase+PCL711_SOFTTRIG);

	for(i=0;i<PCL711_TIMEOUT;i++){
		hi=inb(dev->iobase+PCL711_AD_HI);
		if( !(hi&PCL711_DRDY) )
			break;
		udelay(25);
	}
	if(i==PCL711_TIMEOUT){
		printk("comedi%d: pcl711: A/D timeout\n",dev->minor);
		return -ETIME;
	}
	lo=inb(dev->iobase+PCL711_AD_LO);

	it->data=((hi&0xf)<<8)|lo;
		
	return 0;
}

static int pcl711_ai(comedi_device *dev,comedi_trig *it)
{
	if(dev->busy[0])return -EBUSY;
	switch(it->trigsrc){
	case COMEDI_TRIGEXT:
		if(!devpriv->is_pcl711b || dev->irq==0)return -EINVAL;
		
		dev->busy[0]=1;
		devpriv->ntrig=it->n;
		devpriv->adchan=it->chan;
		outb(devpriv->aip[it->chan],dev->iobase+PCL711_GAIN);
		outb(it->chan,dev->iobase+PCL711_MUX);
		outb((dev->irq<<4)|3,dev->iobase+PCL711_MODE);
		
		break;
	case COMEDI_TRIGCLK:
		if(!devpriv->is_pcl711b || dev->irq==0)return -EINVAL;
		
		dev->busy[0]=1;
		devpriv->ntrig=it->n;
		devpriv->adchan=it->chan;
		outb(devpriv->aip[it->chan],dev->iobase+PCL711_GAIN);
		outb(it->chan,dev->iobase+PCL711_MUX);
		
		outb(0x74,dev->iobase+PCL711_CTRCTL);
		outb((it->trigvar>>16)&0xff,dev->iobase+PCL711_CTR1);
		outb((it->trigvar>>24)&0xff,dev->iobase+PCL711_CTR1);
		outb(0xb4,dev->iobase+PCL711_CTRCTL);
		outb(it->trigvar&0xff,dev->iobase+PCL711_CTR1);
		outb((it->trigvar>>8)&0xff,dev->iobase+PCL711_CTR1);
	
		outb((dev->irq<<4)|3,dev->iobase+PCL711_MODE);
		
		break;
	default:
		return -EINVAL;
	}
	
	return 0;
}

static int pcl711_aisp(comedi_device *dev,comedi_param *it)
{
	switch(it->pnum){
	case COMEDI_RANGE:
		if(it->pval<0 || it->pval>=5)return -EINVAL;
		devpriv->aip[it->chan]=it->pval;
		changeparam(&dev->chinfo[it->chan].paramlist,COMEDI_RANGE,
			0,it->pval);
		break;
	default:
		return -EINVAL;
	}
	return 0;
}

static int pcl711_da(comedi_device *dev,comedi_trig *it)
{
	if(it->data>0xfff)return -EINVAL;

	outb((it->data&0xff),dev->iobase+PCL711_DA_LO);
	outb((it->data>>8),dev->iobase+PCL711_DA_HI);
	
	return 0;
}

static int pcl711_di(comedi_device *dev,comedi_trig *it)
{
	comedi_sample s;
	
	s.data=inb(dev->iobase+PCL711_DI_LO)|
		(inb(dev->iobase+PCL711_DI_HI)<<8);
	it->data=s.data;
	
	return 0;
}

static int pcl711_do(comedi_device *dev,comedi_trig *it)
{
	int mask,data;
	
	if(it->n)mask=it->n; else mask=0xffff;
	data=(devpriv->last_do&~mask)|(it->data&mask);
	outb(data&0xff,dev->iobase+PCL711_DO_LO);
	outb((data>>8),dev->iobase+PCL711_DO_HI);
	devpriv->last_do=data;
	
	return 0;
}

int pcl711_init(comedi_device *dev,comedi_devconfig *it)
{
	int i;
	int is_b;
	int irq;
	
	if(!strcmp("pcl711",it->name))is_b=0;
	else if(!strcmp("pcl711b",it->name))is_b=1;
	else return 0;

	dev->iobase=it->options[0];
	printk("comedi%d: pcl711: 0x%04x ",dev->minor,dev->iobase);
	if(check_region(dev->iobase,PCL711_SIZE)<0){
		printk("I/O port conflict\n");
		return -EIO;
	}
	
	if(is_b) dev->board_name="pcl711b";
	else dev->board_name="pcl711";
	
	dev->driver_name="pcl711";
	
	irq=it->options[1];
	if(!is_b)irq=0;
	if(irq<0 || irq==1 || irq>7){
		printk("irq out of range\n");
		return -EIO;
	}
	if(irq){
		if(request_irq(irq,pcl711_interrupt,SA_INTERRUPT,"pcl711",dev)){
			printk("unable to allocate irq %d\n",irq);
			irq=0;
		}else{
			printk("( irq = %d )\n",irq);
		}
	}
	dev->irq=irq;
	
	request_region(dev->iobase,PCL711_SIZE,"pcl711");
	dev->iosize=PCL711_SIZE;

	dev->chan=11;

	dev->private=kmalloc(sizeof(pcl711_private)+
		dev->chan*sizeof(comedi_chinfo),GFP_KERNEL);
	if(!dev->private){
		return -ENOMEM;
	}
	memset(dev->private,0,sizeof(pcl711_private)+
		dev->chan*sizeof(comedi_chinfo));
	dev->chinfo=dev->private+sizeof(pcl711_private);
	devpriv->is_pcl711b=is_b;
	for(i=0;i<8;i++){
		dev->chinfo[i].itrig=pcl711_iai;
		dev->chinfo[i].trig=pcl711_ai;
		dev->chinfo[i].sp=NULL;
		addparam(&dev->chinfo[i].paramlist,COMEDI_BITS,COMEDI_RDONLY,
			12);
		addparam(&dev->chinfo[i].paramlist,COMEDI_MAXDATA,COMEDI_RDONLY,
			0xfff);
		addparam(&dev->chinfo[i].paramlist,COMEDI_AD,COMEDI_RDONLY,
			COMEDI_AD_ANALOG);
		addparam(&dev->chinfo[i].paramlist,COMEDI_IO,COMEDI_RDONLY,
			COMEDI_IO_INPUT);
		addparam(&dev->chinfo[i].paramlist,COMEDI_LASTDATA,
			COMEDI_RDONLY,0);
		dev->chinfo[i].lastdata=dev->chinfo[i].paramlist;
		addparam(&dev->chinfo[i].paramlist,COMEDI_RANGE,0,0);
		if(is_b){
			addparam(&dev->chinfo[i].paramlist,COMEDI_RANGE_TYPE,
				COMEDI_RDONLY,RANGE_pcl711b_ai);
			dev->chinfo[i].sp=pcl711_aisp;
		}else{
			addparam(&dev->chinfo[i].paramlist,COMEDI_RANGE_TYPE,
				COMEDI_RDONLY,RANGE_bipolar5);
		}
	}

	dev->chinfo[i].itrig=pcl711_da;
	dev->chinfo[i].sp=NULL;
	addparam(&dev->chinfo[i].paramlist,COMEDI_BITS,COMEDI_RDONLY,12);
	addparam(&dev->chinfo[i].paramlist,COMEDI_MAXDATA,COMEDI_RDONLY,0xfff);
	addparam(&dev->chinfo[i].paramlist,COMEDI_AD,COMEDI_RDONLY,
		COMEDI_AD_ANALOG);
	addparam(&dev->chinfo[i].paramlist,COMEDI_IO,COMEDI_RDONLY,
		COMEDI_IO_OUTPUT);
	addparam(&dev->chinfo[i].paramlist,COMEDI_LASTDATA,COMEDI_RDONLY,0);
	addparam(&dev->chinfo[i].paramlist,COMEDI_RANGE_TYPE,
		COMEDI_RDONLY,RANGE_bipolar5);
	dev->chinfo[i].lastdata=dev->chinfo[i].paramlist;
	i++;

	dev->chinfo[i].itrig=pcl711_di;
	dev->chinfo[i].sp=NULL;
	addparam(&dev->chinfo[i].paramlist,COMEDI_BITS,COMEDI_RDONLY,16);
	addparam(&dev->chinfo[i].paramlist,COMEDI_AD,COMEDI_RDONLY,
		COMEDI_AD_DIGITAL);
	addparam(&dev->chinfo[i].paramlist,COMEDI_IO,COMEDI_RDONLY,
		COMEDI_IO_INPUT);
	addparam(&dev->chinfo[i].paramlist,COMEDI_LASTDATA,COMEDI_RDONLY,0);
	dev->chinfo[i].lastdata=dev->chinfo[i].paramlist;
	i++;

	dev->chinfo[i].itrig=pcl711_do;
	dev->chinfo[i].sp=NULL;
	addparam(&dev->chinfo[i].paramlist,COMEDI_BITS,COMEDI_RDONLY,16);
	addparam(&dev->chinfo[i].paramlist,COMEDI_AD,COMEDI_RDONLY,
		COMEDI_AD_DIGITAL);
	addparam(&dev->chinfo[i].paramlist,COMEDI_IO,COMEDI_RDONLY,
		COMEDI_IO_OUTPUT);
	addparam(&dev->chinfo[i].paramlist,COMEDI_LASTDATA,COMEDI_RDONLY,0);
	dev->chinfo[i].lastdata=dev->chinfo[i].paramlist;
	i++;
	
	comedi_initbuf(&dev->rdbuf);

	dev->rem=pcl711_rem;

	/* clear DAC */
	outb(0,dev->iobase+PCL711_DA_LO);
	outb(0,dev->iobase+PCL711_DA_HI);

	printk("\n");
	
	return 1;
}


static int pcl711_rem(comedi_device *dev)
{
	int i;
	
	printk("comedi%d: pcl711: remove\n",dev->minor);
	
	if(dev->irq){
		free_irq(dev->irq,dev);
	}
	for(i=0;i<dev->chan;i++)
		delparamlist(dev->chinfo[i].paramlist);
	kfree(dev->private);

	release_region(dev->iobase,dev->iosize);

	comedi_freebuf(&dev->rdbuf);

	return 0;
}



