/*
    module/rti800.c
    hardware driver for Analog Devices RTI-800/815 board

    COMEDI - Linux Control and Measurement Device Interface
    Copyright (C) 1998 David A. Schleef <ds@stm.lbl.gov>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/mm.h>
#include <linux/malloc.h>
#include <linux/errno.h>
#include <linux/ioport.h>
#include <linux/delay.h>
#include <linux/interrupt.h>
#include <linux/timex.h>
#include <linux/timer.h>
#include <asm/io.h>
#include <comedi_module.h>


#define RTI800_SIZE 16

#define RTI800_CSR 0
#define RTI800_MUXGAIN 1
#define RTI800_CONVERT 2
#define RTI800_ADCLO 3
#define RTI800_ADCHI 4
#define RTI800_DAC0LO 5
#define RTI800_DAC0HI 6
#define RTI800_DAC1LO 7
#define RTI800_DAC1HI 8
#define RTI800_CLRFLAGS 9
#define RTI800_DI 10
#define RTI800_DO 11
#define RTI800_9513A_DATA 12
#define RTI800_9513A_CNTRL 13
#define RTI800_9513A_STATUS 13


/*
 * flags
 */

#define RTI800_BUSY 0x80
#define RTI800_DONE 0x40
#define RTI800_OVERRUN 0x20
#define RTI800_TCR 0x10
#define RTI800_DMA_ENAB 0x08
#define RTI800_INTR_TC 0x04
#define RTI800_INTR_EC 0x02
#define RTI800_INTR_OVRN 0x01

#define Am9513_8BITBUS

#define Am9513_output_control(a)	outb(a,dev->iobase+RTI800_9513A_CNTRL)
#define Am9513_output_data(a)		outb(a,dev->iobase+RTI800_9513A_DATA)
#define Am9513_input_data()		inb(dev->iobase+RTI800_9513A_DATA)
#define Am9513_input_status()		inb(dev->iobase+RTI800_9513A_STATUS)

#include <am9513.h>


comedi_devinit rti800_init;
static int rti800_rem(comedi_device *dev);
#ifdef USE_IRQ
static void rti800_interrupt(int irq,void *dev,struct pt_regs * regs);
#endif

typedef struct{
	int junk;
}rti800_private;
#define devpriv ((rti800_private *)dev->private)


#define RTI800_TIMEOUT 100

static int rti800_iai(comedi_device *dev,comedi_trig *it)
{
	int i,hi,lo,gain;
	unsigned long flags;

	gain=getparamval(dev->chinfo[it->chan].paramlist,COMEDI_RANGE);
	outb(it->chan|(gain<<5),dev->iobase+RTI800_MUXGAIN);
	save_flags(flags);
	cli();
	outb(0,dev->iobase+RTI800_CONVERT);
	for(i=0;i<RTI800_TIMEOUT;i++){
		if(inb(dev->iobase+RTI800_CSR)&RTI800_DONE)
			break;
	}
	lo=inb(dev->iobase+RTI800_ADCLO);
	hi=inb(dev->iobase+RTI800_ADCHI);
	restore_flags(flags);

	it->data=(hi<<8)|lo;

	return 0;
}

static int rti800_ai(comedi_device *dev,comedi_trig *it)
{
	switch(it->trigsrc){
/*
	case COMEDI_TRIGCLK:
		if(dev->busy[0])return -EBUSY;
		devpriv->ntrig=it->n;
		outb(it->chan|DT2814_ENB|(it->trigvar<<5),
			dev->iobase+DT2814_CSR);
		break;
*/
	default:
		return -EINVAL;
	}
	
	return 0;
} 

static int rti800_ao(comedi_device *dev,comedi_trig *it)
{
	switch(it->chan){
	case 16:
		outb(it->data&0xff,dev->iobase+RTI800_DAC0LO);
		outb(it->data>>8,dev->iobase+RTI800_DAC0HI);
		return 0;
	case 17:
		outb(it->data&0xff,dev->iobase+RTI800_DAC1LO);
		outb(it->data>>8,dev->iobase+RTI800_DAC1HI);
		return 0;
	}
	return -EINVAL;
}

static int rti800_di(comedi_device *dev,comedi_trig *it)
{
	it->data=inb(dev->iobase+RTI800_DI);
	
	return 0;
}

static int rti800_do(comedi_device *dev,comedi_trig *it)
{
	outb(it->data^0xff,dev->iobase+RTI800_DO);
	
	return 0;
}

static int rti800_aisp(comedi_device *dev,comedi_param *it)
{
	switch(it->pnum){
	case COMEDI_RANGE:
		if(it->pval<0 || it->pval>3)return -EINVAL;
		changeparam(&dev->chinfo[it->chan].paramlist,COMEDI_RANGE,
			0,it->pval);
		return 0;
	default:
		return -EINVAL;
	}
}

int rti800_init(comedi_device *dev,comedi_devconfig *it)
{
#ifdef USE_IRQ
	int irqs,irq;
	long flags;
#endif
	int i;
	int board;

	board=-1;
	if(!strcmp("rti800",it->name))board=0;
	if(!strcmp("rti815",it->name))board=1;
	if(board<0)return 0;

	dev->iobase=it->options[0];
	printk("comedi%d: rti800: 0x%04x ",dev->minor,dev->iobase);
	if(check_region(dev->iobase,RTI800_SIZE)<0){
		printk("I/O port conflict\n");
		return -EIO;
	}
	request_region(dev->iobase,RTI800_SIZE,"rti800");
	dev->iosize=RTI800_SIZE;

	if(board==0)dev->board_name="rti800";
	else dev->board_name="rti815";
	
	dev->driver_name="rti800";

	if(board==0)dev->chan=16+2;
	else dev->chan=16+2+2;

	dev->private=kmalloc(sizeof(rti800_private)+
		dev->chan*sizeof(comedi_chinfo),GFP_KERNEL);
	if(!dev->private){
		return -ENOMEM;
	}
	memset(dev->private,0,sizeof(rti800_private)+
		dev->chan*sizeof(comedi_chinfo));
	dev->chinfo=dev->private+sizeof(rti800_private);
	for(i=0;i<16;i++){
		dev->chinfo[i].itrig=rti800_iai;
		dev->chinfo[i].trig=rti800_ai;
		dev->chinfo[i].sp=rti800_aisp;
		addparam(&dev->chinfo[i].paramlist,COMEDI_BITS,COMEDI_RDONLY,
			12);
		addparam(&dev->chinfo[i].paramlist,COMEDI_MAXDATA,COMEDI_RDONLY,
			0xfff);
		addparam(&dev->chinfo[i].paramlist,COMEDI_AD,COMEDI_RDONLY,
			COMEDI_AD_ANALOG);
		addparam(&dev->chinfo[i].paramlist,COMEDI_IO,COMEDI_RDONLY,
			COMEDI_IO_INPUT);
		addparam(&dev->chinfo[i].paramlist,COMEDI_LASTDATA,
			COMEDI_RDONLY,0);
		addparam(&dev->chinfo[i].paramlist,COMEDI_RANGE,COMEDI_RDWR,
			0);
		dev->chinfo[i].lastdata=dev->chinfo[i].paramlist;
	}
	
	if(board==1){
		dev->chinfo[i].itrig=rti800_ao;
		dev->chinfo[i].sp=NULL;
		addparam(&dev->chinfo[i].paramlist,
			COMEDI_BITS,COMEDI_RDONLY,12);
		addparam(&dev->chinfo[i].paramlist,
			COMEDI_MAXDATA,COMEDI_RDONLY,0xfff);
		addparam(&dev->chinfo[i].paramlist,
			COMEDI_AD,COMEDI_RDONLY,COMEDI_AD_ANALOG);
		addparam(&dev->chinfo[i].paramlist,
			COMEDI_IO,COMEDI_RDONLY,COMEDI_IO_OUTPUT);
		i++;
	
		dev->chinfo[i].itrig=rti800_ao;
		dev->chinfo[i].sp=NULL;
		addparam(&dev->chinfo[i].paramlist,
			COMEDI_BITS,COMEDI_RDONLY,12);
		addparam(&dev->chinfo[i].paramlist,
			COMEDI_MAXDATA,COMEDI_RDONLY,0xfff);
		addparam(&dev->chinfo[i].paramlist,
			COMEDI_AD,COMEDI_RDONLY,COMEDI_AD_ANALOG);
		addparam(&dev->chinfo[i].paramlist,
			COMEDI_IO,COMEDI_RDONLY,COMEDI_IO_OUTPUT);
		i++;
	}

	dev->chinfo[i].itrig=rti800_di;
	dev->chinfo[i].sp=NULL;
	addparam(&dev->chinfo[i].paramlist,
		COMEDI_BITS,COMEDI_RDONLY,8);
	addparam(&dev->chinfo[i].paramlist,
		COMEDI_AD,COMEDI_RDONLY,COMEDI_AD_DIGITAL);
	addparam(&dev->chinfo[i].paramlist,
		COMEDI_IO,COMEDI_RDONLY,COMEDI_IO_INPUT);
	i++;

	dev->chinfo[i].itrig=rti800_do;
	dev->chinfo[i].sp=NULL;
	addparam(&dev->chinfo[i].paramlist,
		COMEDI_BITS,COMEDI_RDONLY,8);
	addparam(&dev->chinfo[i].paramlist,
		COMEDI_AD,COMEDI_RDONLY,COMEDI_AD_DIGITAL);
	addparam(&dev->chinfo[i].paramlist,
		COMEDI_IO,COMEDI_RDONLY,COMEDI_IO_OUTPUT);
	i++;

/* don't yet know how to deal with counter/timers */
#if 0
	dev->chinfo[i].trig=rti800_timer;
	dev->chinfo[i].sp=NULL;
	addparam(&dev->chinfo[i].paramlist,
		COMEDI_BITS,COMEDI_RDONLY,1);
	addparam(&dev->chinfo[i].paramlist,
		COMEDI_AD,COMEDI_RDONLY,COMEDI_AD_DIGITAL);
	addparam(&dev->chinfo[i].paramlist,
		COMEDI_IO,COMEDI_RDONLY,COMEDI_IO_OUTPUT);
	i++;
#endif
	
	comedi_initbuf(&dev->rdbuf);

	dev->rem=rti800_rem;
	
	printk("\n");

	return 1;
}


static int rti800_rem(comedi_device *dev)
{
	int i;
	
	printk("comedi%d: rti800: remove\n",dev->minor);
	
	for(i=0;i<dev->chan;i++)
		delparamlist(dev->chinfo[i].paramlist);
	kfree(dev->private);

	if(dev->irq){
		free_irq(dev->irq,dev);
	}
	release_region(dev->iobase,dev->iosize);

	comedi_freebuf(&dev->rdbuf);

	return 0;
}



